"""RasterDatasetDriver for zarr data."""

import logging
from functools import partial
from os.path import splitext
from typing import Any, Callable, ClassVar

import xarray as xr
from pydantic import Field

from hydromt._utils.unused_kwargs import _warn_on_unused_kwargs
from hydromt.data_catalog.drivers.base_driver import (
    DRIVER_OPTIONS_DESCRIPTION,
    DriverOptions,
)
from hydromt.data_catalog.drivers.preprocessing import get_preprocessor
from hydromt.data_catalog.drivers.raster.raster_dataset_driver import (
    RasterDatasetDriver,
)
from hydromt.error import NoDataStrategy, exec_nodata_strat
from hydromt.typing import (
    Geom,
    SourceMetadata,
    StrPath,
    TimeRange,
    Variables,
    Zoom,
)

logger = logging.getLogger(__name__)

_ZARR_EXT = ".zarr"
_NETCDF_EXT = [".nc", ".netcdf"]


class RasterXarrayOptions(DriverOptions):
    """Options for RasterXarrayDriver."""

    preprocess: str | None = None
    """Name of preprocessor to apply before merging datasets. Available preprocessors include: round_latlon, to_datetimeindex, remove_duplicates, harmonise_dims. See their docstrings for details."""

    ext_override: str | None = None
    """Override the file extension check and try to read all files as the given extension. Useful when reading zarr files without the .zarr extension."""

    def get_preprocessor(self) -> Callable | None:
        """Get the preprocessor function."""
        if self.preprocess is None:
            return None
        return get_preprocessor(self.preprocess)

    def get_ext_override(self, uris: list[str]) -> str | None:
        """Get the extension override."""
        if not self.ext_override:
            return splitext(uris[0])[-1]
        return self.ext_override


class RasterDatasetXarrayDriver(RasterDatasetDriver):
    """
    Driver for RasterDataset using the xarray library: ``raster_xarray``.

    Supports reading and writing zarr and netcdf files using xarray.
    zarr files will be read using `xr.open_zarr` and netcdf files using
    `xr.open_mfdataset`.

    """

    name = "raster_xarray"
    supports_writing = True
    SUPPORTED_EXTENSIONS: ClassVar[set[str]] = {_ZARR_EXT, *_NETCDF_EXT}
    options: RasterXarrayOptions = Field(
        default_factory=RasterXarrayOptions, description=DRIVER_OPTIONS_DESCRIPTION
    )

    def read(
        self,
        uris: list[str],
        *,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
        kwargs_for_open: dict[str, Any] | None = None,
        mask: Geom | None = None,
        variables: Variables | None = None,
        time_range: TimeRange | None = None,
        zoom: Zoom | None = None,
        chunks: dict[str, Any] | None = None,
        metadata: SourceMetadata | None = None,
    ) -> xr.Dataset:
        """
        Read zarr data to an xarray DataSet.

        Args:
        """
        _warn_on_unused_kwargs(
            self.__class__.__name__,
            {
                "zoom": zoom,
            },
        )

        # Sort out the preprocessor
        preprocessor: Callable | None = self.options.get_preprocessor()

        # Check for the override flag
        first_ext = self.options.get_ext_override(uris)

        # When is zarr, open like a zarr archive
        kwargs_for_open = kwargs_for_open or {}
        if first_ext == _ZARR_EXT:
            opn: Callable = partial(
                xr.open_zarr,
                **self.options.get_kwargs() | kwargs_for_open,
            )
            datasets = []
            for _uri in uris:
                ext = splitext(_uri)[-1]
                if ext != first_ext and not self.options.ext_override:
                    logger.warning(f"Reading zarr and {_uri} was not, skipping...")
                else:
                    datasets.append(
                        preprocessor(opn(_uri)) if preprocessor else opn(_uri)
                    )

            ds: xr.Dataset = xr.merge(datasets)

        # Normal netcdf file(s)
        elif first_ext in _NETCDF_EXT:
            filtered_uris = []
            for _uri in uris:
                ext = splitext(_uri)[-1]
                if ext != first_ext:
                    logger.warning(f"Reading netcdf and {_uri} was not, skipping...")
                else:
                    filtered_uris.append(_uri)
            ds: xr.Dataset = xr.open_mfdataset(
                filtered_uris,
                decode_coords="all",
                preprocess=preprocessor,
                **self.options.get_kwargs() | kwargs_for_open,
            )
        else:
            raise ValueError(
                f"Unknown extension for RasterDatasetXarrayDriver: {first_ext} "
            )
        for variable in ds.data_vars:
            if ds[variable].size == 0:
                exec_nodata_strat(
                    f"No data from driver: '{self.name}' for variable: '{variable}'",
                    strategy=handle_nodata,
                )
        return ds

    def write(self, path: StrPath, ds: xr.Dataset, **kwargs) -> str:
        """
        Write the RasterDataset to a local file using zarr.

        args:

        returns: str with written uri
        """
        no_ext, ext = splitext(path)
        # set filepath if incompat
        if ext not in self.SUPPORTED_EXTENSIONS:
            logger.warning(
                f"Unknown extension for RasterDatasetXarrayDriver: {ext},"
                "switching to zarr"
            )
            path = no_ext + _ZARR_EXT
            ext = _ZARR_EXT
        if ext == _ZARR_EXT:
            ds.to_zarr(path, mode="w", **kwargs)
        else:
            ds.to_netcdf(path, **kwargs)

        return str(path)



"""All drivers that can read RasterDatasets."""

from hydromt.data_catalog.drivers.raster.raster_dataset_driver import (
    RasterDatasetDriver,
)
from hydromt.data_catalog.drivers.raster.raster_xarray_driver import (
    RasterDatasetXarrayDriver,
    RasterXarrayOptions,
)
from hydromt.data_catalog.drivers.raster.rasterio_driver import (
    RasterioDriver,
    RasterioOptions,
)

__all__ = [
    "RasterDatasetDriver",
    "RasterDatasetXarrayDriver",
    "RasterioDriver",
    "RasterXarrayOptions",
    "RasterioOptions",
]



"""HydroMT Model API."""

from hydromt.model.model import Model
from hydromt.model.root import ModelRoot
from hydromt.model.steps import hydromt_step

__all__ = ["Model", "hydromt_step", "ModelRoot"]



"""GeoDatasetVectorDriver class for reading vector data from table like files such as csv or parquet."""

import logging
from typing import Any, ClassVar

import xarray as xr
from pydantic import Field

from hydromt._utils.unused_kwargs import _warn_on_unused_kwargs
from hydromt.data_catalog.drivers.base_driver import (
    DRIVER_OPTIONS_DESCRIPTION,
)
from hydromt.data_catalog.drivers.geodataset.geodataset_driver import (
    GeoDatasetDriver,
    GeoDatasetOptions,
)
from hydromt.error import NoDataStrategy, exec_nodata_strat
from hydromt.io import open_geodataset
from hydromt.typing import CRS, Geom, Predicate, SourceMetadata, StrPath, TimeRange

logger = logging.getLogger(__name__)


class GeoDatasetVectorDriver(GeoDatasetDriver):
    """
    Driver for GeoDataset using hydromt vector: ``geodataframe_vector``.

    Supports reading geodataset from a combination of a geometry file and optionally an
    external data file. The geometry file can be any file supported by
    `geopandas.read_file`, such as shapefile, geojson, geopackage, or a tabular file
    like csv or parquet for points (containing latitude and longitude columns). The
    external data file can be a netcdf file or any tabular file like csv or parquet.
    The geometry file and the external data file can be linked using a common column
    (e.g. an ID column).

    """

    name: ClassVar[str] = "geodataset_vector"
    SUPPORTED_EXTENSIONS: ClassVar[set[str]] = {
        ".csv",
        ".parquet",
        ".xlsx",
        ".xls",
        ".xy",
        ".gpkg",
        ".shp",
        ".geojson",
        ".fgb",
    }

    options: GeoDatasetOptions = Field(
        default_factory=GeoDatasetOptions, description=DRIVER_OPTIONS_DESCRIPTION
    )

    def read(
        self,
        uris: list[str],
        *,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
        kwargs_for_open: dict[str, Any] | None = None,
        mask: Geom | None = None,
        predicate: Predicate = "intersects",
        variables: list[str] | None = None,
        time_range: TimeRange | None = None,
        metadata: SourceMetadata | None = None,
    ) -> xr.Dataset:
        """
        Read tabular datafiles like csv or parquet into to an xarray DataSet.

        Args:
        """
        _warn_on_unused_kwargs(
            self.__class__.__name__,
            {
                "variables": variables,
                "time_range": time_range,
                "metadata": metadata,
            },
        )
        # we want to maintain a list as argument to keep the interface compatible with other drivers.
        if len(uris) > 1:
            raise ValueError(
                "GeodatasetVectorDriver only supports reading from one URI per source"
            )
        else:
            uri = uris[0]

        preprocessor = self.options.get_preprocessor()
        kwargs_for_open = kwargs_for_open or {}
        kwargs = self.options.get_kwargs() | kwargs_for_open
        crs: CRS | None = metadata.crs if metadata else None
        data = open_geodataset(
            loc_path=uri, geom=mask, crs=crs, predicate=predicate, **kwargs
        )

        if preprocessor is None:
            out = data
        else:
            out = preprocessor(data)

        if isinstance(out, xr.DataArray):
            if out.size == 0:
                exec_nodata_strat(
                    f"No data from driver {self}'.", strategy=handle_nodata
                )
                return out.to_dataset()
        else:
            for variable in out.data_vars:
                if out[variable].size == 0:
                    exec_nodata_strat(
                        f"No data from driver {self}' for variable {variable}.",
                        strategy=handle_nodata,
                    )
            return out

    def write(
        self,
        path: StrPath,
        ds: xr.Dataset,
        **kwargs,
    ) -> str:
        """Not implemented."""
        raise NotImplementedError("GeodatasetVectorDriver does not support writing. ")



"""URIResolver using HydroMT naming conventions."""

import logging
from functools import reduce
from itertools import chain, product
from typing import Any, Iterable, Optional

import pandas as pd
from fsspec.core import split_protocol

from hydromt._utils.naming_convention import _expand_uri_placeholders
from hydromt.data_catalog.uri_resolvers.uri_resolver import URIResolver
from hydromt.error import NoDataStrategy, exec_nodata_strat
from hydromt.gis.gis_utils import zoom_to_overview_level
from hydromt.typing import (
    Geom,
    SourceMetadata,
    TimeRange,
    Zoom,
)

logger = logging.getLogger(__name__)


class ConventionResolver(URIResolver):
    """URIDataResolver using HydroMT naming conventions."""

    _uri_placeholders = frozenset(
        {"year", "month", "variable", "name", "overview_level"}
    )
    name = "convention"

    def _get_dates(
        self,
        keys: list[str],
        time_range: TimeRange,
    ) -> pd.PeriodIndex:
        """Obtain the dates the user is searching for."""
        t_range: pd.DatetimeIndex = pd.to_datetime([time_range.start, time_range.end])
        freq: str = "M" if "month" in keys else "Y"
        dates: pd.PeriodIndex = pd.period_range(*t_range, freq=freq)
        return dates

    def _resolve_wildcards(self, uris: Iterable[str]) -> set[str]:
        """Expand on the wildcards in the uris based on the filesystem."""

        def split_and_glob(uri: str) -> tuple[Optional[str], list[str]]:
            protocol, _ = split_protocol(uri)
            return (protocol, self.filesystem.get_fs().glob(uri))

        def maybe_unstrip_protocol(
            pair: tuple[Optional[str], Iterable[str]],
        ) -> Iterable[str]:
            if pair[0] is not None:
                return map(
                    lambda uri: self.filesystem.get_fs().unstrip_protocol(uri)
                    if not uri.startswith(pair[0])
                    else uri,
                    pair[1],
                )
            else:
                return pair[1]

        return set(
            chain.from_iterable(  # flatten result
                map(
                    lambda uri_pair: maybe_unstrip_protocol(
                        uri_pair
                    ),  # keep protocol in uri if present
                    reduce(
                        lambda uri_res, uri: uri_res + [split_and_glob(uri)], uris, []
                    ),
                )
            )
        )

    def resolve(
        self,
        uri: str,
        *,
        time_range: Optional[TimeRange] = None,
        mask: Optional[Geom] = None,
        zoom: Optional[Zoom] = None,
        variables: Optional[list[str]] = None,
        metadata: Optional[SourceMetadata] = None,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
    ) -> list[str]:
        """Resolve the placeholders in the URI using naming conventions.

        Parameters
        ----------
        uri : str
            Unique Resource Identifier
        time_range : Optional[TimeRange], optional
            left-inclusive start end time of the data, by default None
        mask : Optional[Geom], optional
            A geometry defining the area of interest, by default None
        zoom: Optional[Zoom], optional
            zoom of the dataset, by default None
        variables : Optional[list[str]], optional
            Names of variables to return, or all if None, by default None
        metadata: Optional[SourceMetadata], optional
            DataSource metadata.
        handle_nodata : NoDataStrategy, optional
            how to react when no data is found, by default NoDataStrategy.RAISE

        Returns
        -------
        list[str]
            a list of expanded uris

        Raises
        ------
        NoDataException
            when no data is found and `handle_nodata` is `NoDataStrategy.RAISE`
        """
        logger.debug(f"attempting to resolve uri: {uri}")

        if metadata is None:
            metadata: SourceMetadata = SourceMetadata()

        uri_expanded, keys, _ = _expand_uri_placeholders(
            uri,
            placeholders=self._uri_placeholders,
            time_range=time_range,
            variables=variables,
        )
        if time_range:
            dates = self._get_dates(keys, time_range)
        else:
            dates = pd.PeriodIndex(["1970-01-01"], freq="d")  # fill any valid value
        if not variables:
            variables = [""]  # fill any valid value
        if zoom:
            zls_dict: Optional[dict[int, float]] = None
            if metadata is not None and hasattr(metadata, "zls_dict"):
                zls_dict = metadata.zls_dict
            crs = None if metadata is None else metadata.crs
            overview_level: int = zoom_to_overview_level(zoom, mask, zls_dict, crs) or 0
        else:
            overview_level = 0  # fill any valid value

        fmts: Iterable[dict[str, Any]] = map(
            lambda t: {
                "year": t[0].year,
                "month": t[0].month,
                "variable": t[1],
                "overview_level": overview_level,
            },
            product(dates, variables),
        )
        uris: list[str] = list(
            self._resolve_wildcards(map(lambda fmt: uri_expanded.format(**fmt), fmts))
        )
        if not uris:
            exec_nodata_strat(
                f"resolver '{self.name}' found no files.", strategy=handle_nodata
            )
            return []  # if ignore

        return uris



"""Metadata on DataSource."""

from typing import Any, Dict, Optional, Union

from pydantic import BaseModel, BeforeValidator, ConfigDict, Field
from typing_extensions import Annotated

from hydromt.typing.crs import CRS

# always stringify version
Version = Annotated[str, BeforeValidator(str)]


class SourceMetadata(BaseModel):
    """
    Metadata for data source.

    This refers to data that is used to enrich the data format the source is in.
    SourceMetaData is not used to reproject or fill nodata values, it is used to
    check the data and enrich the metadata for HydroMT.


    Only the fields listed here are used in HydroMT, the rest are free for used
    to fill in.
    """

    model_config = ConfigDict(extra="allow")

    crs: Optional[CRS] = None
    unit: Optional[str] = None
    extent: Dict[str, Any] = Field(default_factory=dict)
    nodata: Union[dict, float, int, None] = None
    attrs: Dict[str, Any] = Field(default_factory=dict)
    category: Optional[str] = None



"""BaseModel for DataAdapter."""

from datetime import timedelta
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, ConfigDict, Field

from hydromt.typing import TimeRange


class DataAdapterBase(BaseModel):
    """BaseModel for DataAdapter."""

    model_config = ConfigDict(extra="forbid")

    unit_add: Dict[str, Any] = Field(default_factory=dict)
    unit_mult: Dict[str, Any] = Field(default_factory=dict)
    rename: Dict[str, str] = Field(default_factory=dict)

    def _to_source_timerange(
        self,
        time_range: Optional[TimeRange],
    ) -> Optional[TimeRange]:
        """Transform a DataSource timerange to the source-native timerange.

        Parameters
        ----------
        time_range : Optional[TimeRange]
            start and end datetime

        Returns
        -------
        Optional[TimeRange]
            time_range in source format
        """
        if time_range is None:
            return None
        elif dt := self.unit_add.get("time"):
            # subtract from source unit add
            return TimeRange(
                start=time_range.start - timedelta(seconds=dt),
                end=time_range.end - timedelta(seconds=dt),
            )
        else:
            return time_range

    def _to_source_variables(
        self, variables: Optional[List[str]]
    ) -> Optional[List[str]]:
        """Transform DataSource variables to the source-native names.

        Parameters
        ----------
        variables : Optional[List[str]]
            name(s) of the variables in the data.

        Returns
        -------
        Optional[List[str]]
            _description_
        """
        if variables:
            inverse_rename_mapping: dict[str, str] = {
                v: k for k, v in self.rename.items()
            }
            return [inverse_rename_mapping.get(var, var) for var in variables]



"""Data Adapters are generic for its HydroMT type and perform transformations."""

from hydromt.data_catalog.adapters.data_adapter_base import DataAdapterBase
from hydromt.data_catalog.adapters.dataframe import DataFrameAdapter
from hydromt.data_catalog.adapters.dataset import DatasetAdapter
from hydromt.data_catalog.adapters.geodataframe import GeoDataFrameAdapter
from hydromt.data_catalog.adapters.geodataset import GeoDatasetAdapter
from hydromt.data_catalog.adapters.rasterdataset import RasterDatasetAdapter

__all__ = [
    "DataAdapterBase",
    "GeoDataFrameAdapter",
    "DatasetAdapter",
    "DataFrameAdapter",
    "GeoDatasetAdapter",
    "RasterDatasetAdapter",
]



"""Grid Component."""

import logging
from typing import TYPE_CHECKING, Dict, Optional, Tuple, Union, cast

import geopandas as gpd
import numpy as np
import xarray as xr
from affine import Affine
from pyproj import CRS
from shapely.geometry import box

from hydromt.io.readers import open_ncs
from hydromt.io.writers import write_nc
from hydromt.model.components.base import ModelComponent
from hydromt.model.components.spatial import SpatialModelComponent
from hydromt.model.steps import hydromt_step

if TYPE_CHECKING:
    from hydromt.model.model import Model

__all__ = ["GridComponent"]

logger = logging.getLogger(__name__)


class GridComponent(SpatialModelComponent):
    """ModelComponent class for grid components.

    This class is used for setting, creating, writing, and reading regular grid data for a
    HydroMT model. The grid component data stored in the ``data`` property of this class is of the
    hydromt.gis.raster.RasterDataset type which is an extension of xarray.Dataset for regular grid.
    """

    def __init__(
        self,
        model: "Model",
        *,
        filename: str = "grid/grid.nc",
        region_component: Optional[str] = None,
        region_filename: Optional[str] = "grid/grid_region.geojson",
    ):
        """
        Initialize a GridComponent.

        Parameters
        ----------
        model: Model
            HydroMT model instance
        filename: str
            The path to use for reading and writing of component data by default.
            By default "grid/grid.nc".
        region_component: str, optional
            The name of the region component to use as reference for this component's region.
            If provided, the region is not written to disk.
            If None, the region will be set to the grid extent. Note that the create
            method only works if the region_component is None. For add_data_from_*
            methods, the other region_component should be a reference to another
            grid component for correct reprojection.
        region_filename: Optional[str] = "grid/grid_region.geojson",
            The path to use for writing the region data to a file.
            By default "grid/grid_region.geojson". If None, the region is not written to disk.
        """
        # region_component referencing is not possible for grids. The region should be passed via create().
        super().__init__(
            model=model,
            region_component=region_component,
            region_filename=region_filename,
        )
        self._data: Optional[xr.Dataset] = None
        self._filename: str = filename

    def set(
        self,
        data: Union[xr.DataArray, xr.Dataset, np.ndarray],
        name: Optional[str] = None,
        mask: Optional[Union[str, xr.DataArray]] = None,
        force_sn: bool = False,
    ):
        """Add data to grid.

        All layers of grid must have identical spatial coordinates.

        Parameters
        ----------
        data: xarray.DataArray or xarray.Dataset
            new map layer to add to grid
        name: str, optional
            Name of new map layer, this is used to overwrite the name of a DataArray
            and ignored if data is a Dataset
        mask: xr.DataArray, optional
            Name of the mask layer in the grid (self) or data, or directly the mask layer to use.
            Should be a DataArray where `.raster.nodata` is used to define the mask.
            If None or not present as a layer, no masking is applied.
        force_sn: bool, optional, default=False
            If True, the y-axis is oriented such that increasing y values go from South to North.
            If False, incoming data is used as is.
        """
        self._initialize_grid()
        assert self._data is not None

        name_required = isinstance(data, np.ndarray) or (
            isinstance(data, xr.DataArray) and data.name is None
        )
        if name is None and name_required:
            raise ValueError(f"Unable to set {type(data).__name__} data without a name")
        if isinstance(data, np.ndarray):
            if data.size == 0:
                raise ValueError("Cannot set grid data with empty array")
            if data.shape != self._data.raster.shape:
                raise ValueError("Shape of data and grid maps do not match")
            data = xr.DataArray(dims=self._data.raster.dims, data=data, name=name)

        if isinstance(data, xr.DataArray):
            if name is not None:
                data.name = name
            data = data.to_dataset()

        if not isinstance(data, xr.Dataset):
            raise ValueError(f"cannot set data of type {type(data).__name__}")

        if data.raster.res[1] < 0 and force_sn:
            data = data.raster.flipud()

        # Set the data per layer
        if len(self._data) == 0:  # empty grid
            self._data = data
        else:
            mask = self.get_mask_layer(mask, self.data, data)

            for dvar in data.data_vars:
                if dvar in self._data:
                    logger.warning(f"Replacing grid map: {dvar}")
                if mask is not None:
                    if data[dvar].dtype != np.bool:
                        data[dvar] = data[dvar].where(mask, data[dvar].raster.nodata)
                    else:
                        data[dvar] = data[dvar].where(mask, False)
                self._data[dvar] = data[dvar]

    @hydromt_step
    def write(
        self,
        filename: Optional[str] = None,
        *,
        gdal_compliant: bool = False,
        rename_dims: bool = False,
        force_sn: bool = False,
        **kwargs,
    ) -> None:
        """Write model grid data to netcdf file at <root>/<fn>.

        key-word arguments are passed to :py:meth:`~hydromt.model.Model.write_nc`

        Parameters
        ----------
        filename : str, optional
            filename relative to model root, by default 'grid/grid.nc'
        gdal_compliant : bool, optional
            If True, write grid data in a way that is compatible with GDAL,
            by default False
        rename_dims: bool, optional
            If True and gdal_compliant, rename x_dim and y_dim to standard names
            depending on the CRS (x/y for projected and lat/lon for geographic).
        force_sn: bool, optional
            If True and gdal_compliant, forces the dataset to have
            South -> North orientation.
        **kwargs : dict
            Additional keyword arguments to be passed to the `write_nc` method.
        """
        self.root._assert_write_mode()

        if len(self.data) == 0:
            logger.info(
                f"{self.model.name}.{self.name_in_model}: No grid data found, skip writing."
            )
            return

        filename = filename or self._filename
        full_path = self.root.path / filename
        full_path.parent.mkdir(parents=True, exist_ok=True)
        logger.info(
            f"{self.model.name}.{self.name_in_model}: Writing grid data to {full_path}."
        )

        close_handle = write_nc(
            self.data,
            file_path=full_path,
            gdal_compliant=gdal_compliant,
            rename_dims=rename_dims,
            force_overwrite=self.root.mode.is_override_mode(),
            force_sn=force_sn,
            **kwargs,
        )
        if close_handle is not None:
            self._deferred_file_close_handles.append(close_handle)

    @hydromt_step
    def read(self, filename: Optional[str] = None, **kwargs) -> None:
        """Read model grid data at <root>/<fn> and add to grid property.

        key-word arguments are passed to :py:meth:`~hydromt.model.Model.open_nc`

        Parameters
        ----------
        filename : str, optional
            filename relative to model root, by default 'grid/grid.nc'
        **kwargs : dict
            Additional keyword arguments to be passed to the `open_nc` method.
        """
        self.root._assert_read_mode()
        self._initialize_grid(skip_read=True)

        loaded_nc_files = open_ncs(
            filename or self._filename,
            self.root.path,
            **kwargs,
        )
        for ds in loaded_nc_files.values():
            self.set(ds)
            self._open_datasets.append(ds)

    @property
    def res(self) -> Optional[Tuple[float, float]]:
        """Returns the resolution of the model grid."""
        if len(self.data) > 0:
            return self.data.raster.res
        logger.warning("No grid data found for deriving resolution")
        return None

    @property
    def transform(self) -> Optional[Affine]:
        """Returns spatial transform of the model grid."""
        if len(self.data) > 0:
            return self.data.raster.transform
        logger.warning("No grid data found for deriving transform")
        return None

    @property
    def crs(self) -> Optional[CRS]:
        """Returns coordinate reference system embedded in the model grid."""
        if self.data.raster is None:
            logger.warning("No grid data found for deriving crs")
            return None
        if self.data.raster.crs is None:
            logger.warning("No crs found in grid data")
            return None
        return CRS(self.data.raster.crs)

    @property
    def bounds(self) -> Optional[Tuple[float, float, float, float]]:
        """Returns the bounding box of the model grid."""
        if len(self.data) > 0:
            return self.data.raster.bounds
        logger.warning("No grid data found for deriving bounds")
        return None

    @property
    def _region_data(self) -> Optional[gpd.GeoDataFrame]:
        """Returns the geometry of the model area of interest."""
        if len(self.data) > 0:
            crs: Optional[Union[int, CRS]] = self.crs
            if crs is not None and hasattr(crs, "to_epsg"):
                crs = crs.to_epsg()  # not all CRS have an EPSG code
            return gpd.GeoDataFrame(geometry=[box(*self.bounds)], crs=crs)
        logger.warning("No grid data found for deriving region")
        return None

    @property
    def data(self) -> xr.Dataset:
        """Model static gridded data as xarray.Dataset."""
        if self._data is None:
            self._initialize_grid()
        assert self._data is not None
        return self._data

    def _initialize_grid(self, skip_read: bool = False) -> None:
        """Initialize grid object."""
        if self._data is None:
            self._data = xr.Dataset()
            if self.root.is_reading_mode() and not skip_read:
                self.read()

    def test_equal(self, other: ModelComponent) -> Tuple[bool, Dict[str, str]]:
        """Test if two components are equal.

        Parameters
        ----------
        other : ModelComponent
            The component to compare against.

        Returns
        -------
        Tuple[bool, Dict[str, str]]
            True if the components are equal, and a dict with the associated errors per property checked.
        """
        eq, errors = super().test_equal(other)
        if not eq:
            return eq, errors
        other_grid = cast(GridComponent, other)
        try:
            xr.testing.assert_allclose(self.data, other_grid.data)
        except AssertionError as e:
            errors["data"] = str(e)

        return len(errors) == 0, errors

    def _get_grid_data(self) -> Union[xr.DataArray, xr.Dataset]:
        """Get grid data as xarray.DataArray from this component or the reference."""
        if self._region_component is not None:
            reference_component = self.model.get_component(self._region_component)
            if not isinstance(reference_component, GridComponent):
                raise ValueError(
                    f"Unable to find the referenced grid component: '{self._region_component}'."
                )
            if reference_component.data is None:
                raise ValueError(
                    f"Unable to get grid from the referenced region component: '{self._region_component}'."
                )
            return reference_component.data

        if self.data is None:
            raise ValueError("Unable to get grid data from this component.")
        return self.data

    @staticmethod
    def get_mask_layer(mask: str | xr.DataArray | None, *args) -> xr.DataArray | None:
        """Get the proper mask layer based on itself or a layer in a Dataset.

        Parameters
        ----------
        mask : str | xr.DataArray | None
            The mask itself or the name of the mask layer in another dataset.
        *args : list
            These have to be xarray Datasets in which the mask (as a string)
            can be present
        """
        if mask is None:
            return None
        if isinstance(mask, xr.DataArray):
            return mask != mask.raster.nodata
        if not isinstance(mask, str):
            raise ValueError(
                f"Unknown type for determining mask: {type(mask).__name__}"
            )
        for ds in args:
            if mask in ds:
                return ds[mask] != ds[mask].raster.nodata
        return None  # Nothin found



"""Driver for GeoDataFrames."""

import logging
from abc import ABC, abstractmethod
from typing import Any

import geopandas as gpd

from hydromt.data_catalog.drivers import BaseDriver
from hydromt.error import NoDataStrategy
from hydromt.typing import SourceMetadata, StrPath

logger = logging.getLogger(__name__)


class GeoDataFrameDriver(BaseDriver, ABC):
    """Abstract Driver to read GeoDataFrames."""

    @abstractmethod
    def read(
        self,
        uris: list[str],
        *,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
        kwargs_for_open: dict[str, Any] | None = None,
        metadata: SourceMetadata | None = None,
        mask: Any = None,
        predicate: str = "intersects",
        variables: str | list[str] | None = None,
    ) -> gpd.GeoDataFrame:
        """Read in any compatible data source to a geopandas `GeoDataFrame`."""
        ...

    def write(
        self,
        path: StrPath,
        gdf: gpd.GeoDataFrame,
        **kwargs,
    ) -> str:
        """
        Write out a GeoDataFrame to file.

        Not all drivers should have a write function, so this method is not
        abstract.

        args:
        """
        raise NotImplementedError(
            f"Writing using driver '{self.name}' is not supported."
        )



"""HydroMT: Automated and reproducible model building and analysis."""

# version number without 'v' at start
__version__ = "1.3.0-rc3"

# This is only here to suppress the bug described in
# https://github.com/pydata/xarray/issues/7259
# We have to make sure that netcdf4 is imported before
# numpy is imported for the first time, e.g. also via
# importing xarray

import warnings  # noqa: F401

import netCDF4  # noqa: F401

# submodules
from hydromt import data_catalog, gis, io, model, stats

# high-level methods
from hydromt._utils.log import initialize_logging
from hydromt.data_catalog import DataCatalog
from hydromt.gis import raster, vector
from hydromt.model import Model, hydromt_step
from hydromt.plugins import PLUGINS

initialize_logging()

__all__ = [
    # high-level classes
    "DataCatalog",
    "Model",
    # submodules
    "data_catalog",
    "gis",
    "io",
    "model",
    "stats",
    # raster and vector accessor
    "raster",
    "vector",
    # high-level functions
    "hydromt_step",
    # plugins
    "PLUGINS",
]



"""All the definitions of type aliases used in HydroMT."""

from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Literal, Tuple, TypedDict, Union

import geopandas as gpd
import numpy as np
import pandas as pd
from dateutil.parser import parse
from pydantic import (
    BaseModel,
    Field,
    field_serializer,
    field_validator,
    model_validator,
)
from shapely.geometry.base import BaseGeometry
from typing_extensions import Annotated
from xarray import DataArray, Dataset

from hydromt.typing.model_mode import ModelMode


def _validate_bbox(
    bbox: tuple[float, float, float, float],
) -> tuple[float, float, float, float]:
    assert bbox[0] < bbox[2], (
        f"bbox minx: '{bbox[0]}' should be less than maxx: '{bbox[2]}'."
    )
    assert bbox[1] < bbox[3], (
        f"bbox miny: '{bbox[1]}' should be less than maxy: '{bbox[3]}'."
    )
    return bbox


DataType = Literal[
    "DataFrame", "DataSet", "GeoDataFrame", "GeoDataSet", "RasterDataset"
]
Bbox = Annotated[Tuple[float, float, float, float], _validate_bbox]


StrPath = Union[str, Path]
GeoDataframeSource = StrPath
GeoDatasetSource = StrPath
RasterDatasetSource = StrPath
DatasetSource = StrPath

Crs = int
TotalBounds = Tuple[Bbox, Crs]
Zoom = Union[int, Tuple[float, str]]  # level OR (resolution, unit)
Number = Union[int, float]
SourceSpecDict = TypedDict(
    "SourceSpecDict", {"source": str, "provider": str, "version": Union[str, int]}
)

XArrayDict = Dict[str, Union[DataArray, Dataset]]

ExportConfigDict = TypedDict(
    "ExportConfigDict",
    {"args": Dict[str, Any], "meta": Dict[str, Any], "sources": List[SourceSpecDict]},
)

Predicate = Literal[
    "intersects", "within", "contains", "overlaps", "crosses", "touches"
]

Geom = Union[gpd.GeoDataFrame, gpd.GeoSeries]
GpdShapeGeom = Union[Geom, BaseGeometry]

Data = Union[Dataset, DataArray]

Variables = Union[str, List[str]]

GeomBuffer = int

ModeLike = Union[ModelMode, str]

DATETIME_FORMAT: str = "%Y-%m-%d_%H:%M:%S"


class TimeRange(BaseModel):
    """A time range with start and end datetime."""

    start: datetime = Field(..., description="Start of the time range.")
    end: datetime = Field(..., description="End of the time range.")

    @field_validator("start", "end", mode="before")
    def parse_datetime(cls, v: Union[str, datetime]) -> datetime:
        """Parse a datetime from string or numpy datetime64."""
        if isinstance(v, str):
            try:
                # first try to parse as exact format
                return datetime.strptime(v, DATETIME_FORMAT)
            except ValueError:
                # fallback to more flexible parsing
                return parse(v)
        elif isinstance(v, np.datetime64):
            return pd.to_datetime(v).to_pydatetime()
        return v

    @model_validator(mode="after")
    def validate(self) -> "TimeRange":
        """Validate the time range."""
        if self.start > self.end:
            raise ValueError(
                f"time range start: '{self.start}' should be less than end: '{self.end}'"
            )
        return self

    @field_serializer("start", "end", mode="plain")
    def serialize_datetime(self, v: datetime) -> str:
        """Serialize a datetime to string."""
        return v.strftime(DATETIME_FORMAT)

    @staticmethod
    def create(data: Any) -> "TimeRange":
        """Create a TimeRange from various input types."""
        if isinstance(data, TimeRange):
            return data
        elif isinstance(data, dict):
            return TimeRange(**data)
        elif isinstance(data, (list, tuple)) and len(data) == 2:
            return TimeRange(start=data[0], end=data[1])
        else:
            raise ValueError(
                f"Cannot create TimeRange from data: '{data}' of type '{type(data)}'."
            )



"""Script to update the registry.txt file for each predefined catalog."""

from pathlib import Path

from hydromt.data_catalog.predefined_catalog import create_registry_file

if __name__ == "__main__":
    root = Path(__file__).parent
    cat_roots = [d for d in root.iterdir() if d.is_dir()]
    for cat_root in cat_roots:
        create_registry_file(cat_root)



# -*- coding: utf-8 -*-
"""General and basic API for models in HydroMT."""

import logging
import os
import shutil
import typing
from abc import ABCMeta
from inspect import _empty, signature
from os.path import isabs, isfile, join
from pathlib import Path
from typing import Any, Dict, List, Optional, TypeVar, Union, cast

import geopandas as gdp
from pyproj import CRS

from hydromt._utils import _rgetattr, _validate_steps, log
from hydromt.data_catalog import DataCatalog
from hydromt.io.readers import read_yaml
from hydromt.model.components import (
    ModelComponent,
    SpatialModelComponent,
)
from hydromt.model.root import ModelRoot
from hydromt.model.steps import hydromt_step
from hydromt.plugins import PLUGINS

__all__ = ["Model"]

# define hydromt model entry points
# see also hydromt.model group in pyproject.toml
__hydromt_eps__ = ["Model"]


logger = logging.getLogger(__name__)
T = TypeVar("T", bound=ModelComponent)


class Model(object, metaclass=ABCMeta):
    """General and basic API for models in HydroMT."""

    name: str = "model"
    # supported model version should be filled by the plugins
    # e.g. _MODEL_VERSION = ">=1.0, <1.1"
    _MODEL_VERSION = None

    def __init__(
        self,
        root: str | Path | None = None,
        *,
        components: Optional[Dict[str, Any]] = None,
        mode: str = "w",
        data_libs: Optional[Union[List, str]] = None,
        region_component: Optional[str] = None,
        **catalog_keys,
    ):
        """Initialize a model.

        Parameters
        ----------
        root : str, Path, optional
            Model root, by default the current working directory: ``Path.cwd()``
        components: Dict[str, Any], optional
            Dictionary of components to add to the model, by default None
            Every entry in this dictionary contains the name of the component as key,
            and the component object as value, or a dictionary with options passed to the component initializers.
            If a component is a dictionary, the key 'type' should be provided with the
            name of the component type.

            .. code-block:: python

                {
                    "grid": {
                        "type": "GridComponent",
                        "filename": "path/to/grid.nc"
                    }
                }

        mode : {'r','r+','w'}, optional
            read/append/write mode, by default "w"
        data_libs : List[str], optional
            List of data catalog configuration files, by default None
        region_component : str, optional
            The name of the region component in the components dictionary.
            If None, the model will can automatically determine the region component if there is only one `SpatialModelComponent`.
            Otherwise it will raise an error.
            If there are no `SpatialModelComponent` it will raise a warning that `region` functionality will not work.
        **catalog_keys:
            Additional keyword arguments to be passed down to the DataCatalog.
        """
        # Recursively update the options with any defaults that are missing in the configuration.
        components = components or {}

        data_libs = data_libs or []

        # link to data
        self.data_catalog = DataCatalog(data_libs=data_libs, **catalog_keys)
        """DataCatalog for data access"""

        # file system
        path = Path(root) if root is not None else Path.cwd()
        self.root: ModelRoot = ModelRoot(path, mode=mode)
        """Model root"""

        self.components: Dict[str, ModelComponent] = {}
        self._add_components(components)

        model_metadata = cast(Dict[str, str], PLUGINS.model_metadata[self.name])
        logger.info(
            f"Initializing {self.name} model from {model_metadata['plugin_name']} (v{model_metadata['version']})."
        )

        self._region_component_name = self._determine_region_component(region_component)

    def _determine_region_component(
        self, region_component: Optional[str]
    ) -> str | None:
        if region_component is not None:
            if region_component not in self.components:
                raise KeyError(f"Component {region_component} not found in components.")
            elif not isinstance(
                self.components.get(region_component, None), SpatialModelComponent
            ):
                raise ValueError(
                    f"Component {region_component} is not a {SpatialModelComponent.__name__}."
                )
            return region_component
        else:
            has_region_components = [
                (name, c)
                for name, c in self.components.items()
                if isinstance(c, SpatialModelComponent)
            ]
            if len(has_region_components) > 1:
                raise ValueError(
                    "Multiple region components found in components. "
                    "Specify region_component."
                )
            if len(has_region_components) == 0:
                logger.warning("No region component found in components.")
                return None
            return has_region_components[0][0]

    def _add_components(self, components: Dict[str, Any]) -> None:
        """Add all components that are specified in the config file."""
        for name, value in components.items():
            if isinstance(value, ModelComponent):
                self.add_component(name, value)
            elif isinstance(value, dict):
                type_name = value.pop("type")
                component_type = PLUGINS.component_plugins[type_name]
                self.add_component(name, component_type(self, **value))
            else:
                raise ValueError(
                    "Error in components argument. Type is not a ModelComponent or dict."
                )

    def add_component(self, name: str, component: ModelComponent) -> None:
        """Add a component to the model. Will raise an error if the component already exists."""
        if name in self.components:
            raise ValueError(f"Component {name} already exists in the model.")
        if not name.isidentifier():
            raise ValueError(f"Component name {name} is not a valid identifier.")
        self.components[name] = component

    def get_component(self, name: str) -> ModelComponent:
        """Get a component from the model. Will raise an error if the component does not exist."""
        return self.components[name]

    def __getattr__(self, name: str) -> ModelComponent:
        """Get a component from the model. Will raise an error if the component does not exist."""
        return self.get_component(name)

    @staticmethod
    def from_dict(model_dict: Dict[str, Any]) -> "Model":
        """Construct a model with the components and other init arguments in the yaml file located at `path`."""
        model_type = model_dict.pop("modeltype", "Model")
        model_config = model_dict.pop("global", {})
        if model_type not in PLUGINS.model_plugins:
            raise ValueError(
                f"Unknown model type: {model_type}, options are: [{', '.join(PLUGINS.model_plugins)}]"
            )
        else:
            return cast("Model", PLUGINS.model_plugins[model_type](**model_config))

    @staticmethod
    def from_yml(path: Path) -> "Model":
        """Construct a model with the components and other init arguments in the yaml file located at `path`."""
        file_contents = read_yaml(path)
        return Model.from_dict(file_contents)

    @property
    def region(self) -> Optional[gdp.GeoDataFrame]:
        """Return the model's region component."""
        return (
            cast(
                SpatialModelComponent, self.components[self._region_component_name]
            ).region
            if self._region_component_name in self.components
            else None
        )

    @property
    def crs(self) -> Optional[CRS]:
        """Returns coordinate reference system embedded in region."""
        return self.region.crs if self.region is not None else None

    def __enter__(self) -> "Model":
        """Enter the model runtime context.

        This allows the model to be used as a context manager in a ``with`` block.
        The method returns the model instance itself, enabling setup and write
        operations to be performed within the block.

        Example
        -------
        >>> from hydromt import Model
        >>> with Model(root="path/to/model") as model: # This is where __enter__ is called
        ...     model.setup_x()
        ...     model.setup_y()
        ...     model.write()
        # Exiting the with block will call __exit__, ensuring resources are cleaned up.

        Returns
        -------
        Model
            The model instance.
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """Exit the model runtime context.

        Ensures that resources associated with the model are properly released
        when leaving a ``with`` block, even if an exception occurred during its
        execution. This method delegates cleanup to :meth:`close`.

        Example
        -------
        >>> from hydromt import Model
        >>> with Model(root="path/to/model") as model: # This is where __enter__ is called
        ...     model.setup_x()
        ...     model.setup_y()
        ...     model.write()
        # Exiting the with block will call __exit__, ensuring resources are cleaned up.

        Parameters
        ----------
        exc_type : type or None
            The type of the exception raised in the ``with`` block, or ``None``
            if no exception occurred.
        exc_value : BaseException or None
            The exception instance raised in the ``with`` block, or ``None``.
        traceback : types.TracebackType or None
            The traceback object associated with the exception, or ``None``.
        """
        self.close()

    def build(
        self,
        *,
        write: Optional[bool] = True,
        steps: List[Dict[str, Dict[str, Any]]],
    ):
        r"""Single method to build a model from scratch based on settings in `steps`.

        Methods will be run one by one based on the /order of appearance in `steps`
        (configuration file). For a list of available functions see
        :ref:`The model API<model_api>` and
        :ref:`The plugin documentation<plugin_create>`.

        By default the full model will be written at the end, except if a write step
        is called for somewhere in steps, then this is skipped.

        Note that the \* in the signature signifies that all of the arguments to this
        function MUST be provided as keyword arguments.

        Parameters
        ----------
        write: bool, optional
            Write complete model after executing all methods in opt, by default True.
        steps: Optional[List[Dict[str, Dict[str, Any]]]]
            Model build steps. The steps can be parsed from a hydromt workflow/
            configuration file using :py:meth:`~hydromt.io.read_workflow_yaml`.
            This is a list of nested dictionary where the first-level keys are the names
            of the method for a ``Model`` method (e.g. `write`) OR the name of a
            component followed by the name of the method to run separated by a dot for
            ``ModelComponent`` method (e.g. `grid.write`).
            Any subsequent pairs will be passed to the method as arguments.

            .. code-block:: text

                [
                    - <component_name>.<name of method1>: {
                        <argument1>: <value1>, <argument2>: <value2>
                    },
                    - <component_name>.<name of method2>: {
                        ...
                    }
                ]

        """
        with log.to_file(Path(self.root.path) / "hydromt.log"):
            steps = steps or []
            _validate_steps(self, steps)

            for step_dict in steps:
                step, kwargs = next(iter(step_dict.items()))
                logger.info(f"build: {step}")
                # Call the methods.
                method = _rgetattr(self, step)
                params = {
                    param: arg.default
                    for param, arg in signature(method).parameters.items()
                    if arg.default != _empty
                }
                merged: dict[str, Any] = {}
                if params:
                    merged.update(**params)
                if kwargs:
                    merged.update(**kwargs)
                for k, v in merged.items():
                    logger.info(f"{step}.{k}={v}")

                method(**merged)

            # If there are any write options included in the steps,
            # we don't need to write the whole model.
            if write and not self._options_contain_write(steps):
                self.write()

            self.close()

    def update(
        self,
        *,
        model_out: str | Path | None = None,
        write: Optional[bool] = True,
        steps: Optional[List[Dict[str, Dict[str, Any]]]] = None,
        forceful_overwrite: bool = False,
    ):
        r"""Single method to update a model based the settings in `steps`.

        Methods will be run one by one based on the /order of appearance in `steps`
        (configuration file). For a list of available functions see :ref:`The model API<model_api>`
        and :ref:`The plugin documentation<plugin_create>`

        Note that the \* in the signature signifies that all of the arguments to this function
        MUST be provided as keyword arguments.


        Parameters
        ----------
        model_out: str, path, optional
            Destination folder to write the model schematization after updating
            the model. If None the updated model components are overwritten in the
            current model schematization if these exist. By default None.
        write: bool, optional
            Write the updated model schematization to disk. By default True.
        steps: Optional[List[Dict[str, Dict[str, Any]]]]
            Model build steps. The steps can be parsed from a hydromt workflow/
            configuration file using :py:meth:`~hydromt.io.read_workflow_yaml`.
            This is a list of nested dictionary where the first-level keys are the names
            of a component followed by the name of the method to run separated by a dot.
            any subsequent pairs will be passed to the method as arguments.

            .. code-block:: text

                [
                    - <component_name>.<name of method1>: {
                        <argument1>: <value1>, <argument2>: <value2>
                    },
                    - <component_name>.<name of method2>: {
                        ...
                    }
                ]
          forceful_overwrite:
            Force open files to close when attempting to write them. In the case you
            try to write to a file that's already opened. The output will be written
            to a temporary file in case the original file cannot be written to.
        """
        # read current model & optionally clear log file
        with log.to_file(
            self.root.path / "hydromt.log",
            append=self.root.is_reading_mode(),
        ):
            steps = steps or []
            _validate_steps(self, steps)
            if not self.root.is_writing_mode():
                if model_out is None:
                    raise ValueError(
                        '"model_out" directory required when updating in "read-only" mode.'
                    )
                # Copy log file
                model_out = Path(model_out)
                if model_out != self.root.path:
                    model_out.mkdir(parents=True, exist_ok=True)
                    shutil.copyfile(
                        self.root.path / "hydromt.log", model_out / "hydromt.log"
                    )

                self.read()
                mode = "w+" if forceful_overwrite else "w"
                self.root.set(model_out, mode=mode)

        # No need to clear the log file here since we did that already if needed above
        with log.to_file(self.root.path / "hydromt.log", append=True):
            # check if model has a region
            if self._region_component_name is not None and self.region is None:
                raise ValueError(
                    "Model region not found, setup model using `build` first."
                )

            # loop over methods from config file
            for step_dict in steps:
                step, kwargs = next(iter(step_dict.items()))
                logger.info(f"update: {step}")
                # Call the methods.
                method = _rgetattr(self, step)
                params = {
                    param: arg.default
                    for param, arg in signature(method).parameters.items()
                    if arg.default != _empty
                }
                merged = {}
                if params:
                    merged.update(**params)
                if kwargs:
                    merged.update(**kwargs)
                for k, v in merged.items():
                    logger.info(f"{step}.{k}={v}")
                method(**merged)

            # If there are any write options included in the steps,
            # we don't need to write the whole model.
            if write and not self._options_contain_write(steps):
                self.write()

            self.close()

    def close(self) -> None:
        """Close all components by closing their open files."""
        for c in self.components.values():
            c.close()

        # Finish all deferred file closes
        for c in self.components.values():
            c.finish_write()

    @hydromt_step
    def write(self, components: Optional[List[str]] = None) -> None:
        """Write provided components to disk with defaults.

        Parameters
        ----------
            components: Optional[List[str]]
                the components that should be writen to disk. If None is provided
                all components will be written.
        """
        components = components or list(self.components.keys())
        for c in [self.components[name] for name in components]:
            c.write()

    @hydromt_step
    def read(self, components: Optional[List[str]] = None) -> None:
        """Read provided components from disk.

        Parameters
        ----------
            components: Optional[List[str]]
                the components that should be read from disk. If None is provided
                all components will be read.
        """
        logger.info(f"Reading model data from {self.root.path}")
        components = components or list(self.components.keys())
        for c in [self.components[name] for name in components]:
            c.read()

    @staticmethod
    def _options_contain_write(steps: List[Dict[str, Dict[str, Any]]]) -> bool:
        return any("write" in next(iter(step_dict)) for step_dict in steps)

    @hydromt_step
    def write_data_catalog(
        self,
        root: str | Path | None = None,
        data_lib_path: str | Path = "hydromt_data.yml",
        used_only: bool = True,
        append: bool = True,
        save_csv: bool = False,
    ):
        """Write the data catalog to data_lib_path.

        Parameters
        ----------
        root: str, Path, optional
            Global root for all relative paths in configuration file.
            If "auto" the data source paths are relative to the yaml output ``path``.
        data_lib_path: str, Path, optional
            Path of output yml file, absolute or relative to the model root,
            by default "hydromt_data.yml".
        used_only: bool, optional
            If True, export only data entries kept in used_data list. By default True
        append: bool, optional
            If True, append to an existing
        save_csv: bool, optional
            If True, save the data catalog also as an csv table. By default False.
        """
        self.root._assert_write_mode()
        path = (
            data_lib_path
            if isabs(data_lib_path)
            else join(self.root.path, data_lib_path)
        )
        cat = DataCatalog(fallback_lib=None)
        # read hydromt_data yml file and add to data catalog
        if self.root.is_reading_mode() and isfile(path) and append:
            cat.from_yml(path)
        # update data catalog with new used sources
        for name, source in self.data_catalog.list_sources(used_only=used_only):
            # don't have to re add sources we already know
            if source.name in cat.get_source_names():
                continue
            cat.add_source(name, source)
        # write data catalog
        if cat.sources:
            if save_csv:
                csv_path = os.path.splitext(path)[0] + ".csv"
                cat._to_dataframe().reset_index().to_csv(
                    csv_path, sep=",", index=False, header=True
                )

            cat.to_yml(path, root=root)

    def test_equal(self, other: "Model") -> tuple[bool, Dict[str, str]]:
        """Test if two models are equal, based on their components.

        Parameters
        ----------
        other : Model
            The model to compare against.

        Returns
        -------
        Tuple[bool, Dict[str, str]]
            True if equal, dictionary with errors per model component which is not equal.
        """
        if not isinstance(other, self.__class__):
            return False, {
                "__class__": f"f{other.__class__} does not inherit from {self.__class__}."
            }
        components = sorted(list(self.components.keys()))
        components_other = sorted(list(other.components.keys()))
        if components != components_other:
            return False, {
                "components": f"Components do not match: {components} != {components_other}"
            }

        errors: Dict[str, str] = {}
        is_equal = True
        for name, c in self.components.items():
            component_equal, component_errors = c.test_equal(other.components[name])
            is_equal &= component_equal
            errors.update(**component_errors)
        return is_equal, errors


def _assert_isinstance(obj: Any, dtype: Any, name: str = ""):
    """Check if obj match typing or class (dtype)."""
    args = typing.get_args(dtype)
    _cls = typing.get_origin(dtype)
    if len(args) == 0 and dtype != Any and dtype is not None:
        assert isinstance(obj, dtype), name
    elif _cls == Union:
        assert isinstance(obj, args), name
    elif _cls is not None:
        assert isinstance(obj, _cls), name
    # recursive check of dtype dict keys and values
    if len(args) > 0 and _cls is dict:
        for key, val in obj.items():
            _assert_isinstance(key, args[0], f"{name}.{str(key)}")
            _assert_isinstance(val, args[1], f"{name}.{str(key)}")



"""Drivers for tabular data."""

from hydromt.data_catalog.drivers.dataframe.dataframe_driver import DataFrameDriver
from hydromt.data_catalog.drivers.dataframe.pandas_driver import PandasDriver

__all__ = ["DataFrameDriver", "PandasDriver"]



"""Tables component."""

import glob
import logging
from os.path import basename, join
from typing import TYPE_CHECKING, Dict, Optional, Union, cast

import pandas as pd

from hydromt.model.components.base import ModelComponent
from hydromt.model.steps import hydromt_step

if TYPE_CHECKING:
    from hydromt.model.model import Model

__all__ = ["TablesComponent"]

logger = logging.getLogger(__name__)


class TablesComponent(ModelComponent):
    """TablesComponent contains data as a dictionary of pandas.DataFrame.

    It is well suited to represent non-geospatial tabular model data.
    """

    def __init__(
        self,
        model: "Model",
        filename: str = "tables/{name}.csv",
    ):
        """Initialize a TablesComponent.

        Parameters
        ----------
        model: Model
            HydroMT model instance
        filename: str
            The default place that should be used for reading and writing unless the
            user overrides it. If a relative path is given it will be used as being
            relative to the model root. By default `tables/{name}.csv` for this
            component, and can be either relative or absolute.
        """
        self._data: Optional[Dict[str, Union[pd.DataFrame, pd.Series]]] = None
        self._filename: str = filename
        super().__init__(model=model)

    @property
    def data(self) -> Dict[str, Union[pd.DataFrame, pd.Series]]:
        """Model tables."""
        if self._data is None:
            self._initialize_tables()

        assert self._data is not None
        return self._data

    def _initialize_tables(self, skip_read=False) -> None:
        """Initialize the model tables."""
        if self._data is None:
            self._data = dict()
            if self.root.is_reading_mode() and not skip_read:
                self.read()

    @hydromt_step
    def write(self, filename: Optional[str] = None, **kwargs) -> None:
        """Write tables at provided or default file path if none is provided."""
        self.root._assert_write_mode()

        if len(self.data) == 0:
            logger.info(
                f"{self.model.name}.{self.name_in_model}: No tables found, skip writing."
            )
            return

        filename = filename or self._filename

        kwargs.setdefault("index", False)
        kwargs.setdefault("header", True)
        kwargs.setdefault("sep", ",")

        for name, value in self.data.items():
            write_path = self.root.path / filename.format(name=name)
            write_path.parent.mkdir(parents=True, exist_ok=True)
            logger.info(
                f"{self.model.name}.{self.name_in_model}: Writing table to {write_path}."
            )
            value.to_csv(write_path, **kwargs)

    @hydromt_step
    def read(self, filename: Optional[str] = None, **kwargs) -> None:
        """Read tables at provided or default file path if none is provided."""
        self.root._assert_read_mode()
        self._initialize_tables(skip_read=True)
        logger.info("Reading model table files.")
        fn = filename or self._filename
        filenames = glob.glob(join(self.root.path, fn.format(name="*")))
        if len(filenames) > 0:
            for fn in filenames:
                name = basename(fn).split(".")[0]
                tbl = pd.read_csv(fn, **kwargs)
                self.set(tbl, name=name)

    def set(
        self,
        tables: Union[
            Union[pd.DataFrame, pd.Series], Dict[str, Union[pd.DataFrame, pd.Series]]
        ],
        name: Optional[str] = None,
    ) -> None:
        """Add (a) table(s) <pandas.DataFrame> to model.

        Parameters
        ----------
        tables : pandas.DataFrame, pandas.Series or dict
            Table(s) to add to model.
            Multiple tables can be added at once by passing a dict of tables.
        name : str, optional
            Name of table, by default None. Required when tables is not a dict.
        """
        self._initialize_tables()
        assert self._data is not None
        # added here so we only have to declare types once
        tables_to_add: Dict[str, Union[pd.DataFrame, pd.Series]]

        if not isinstance(tables, dict):
            if name is None:
                raise ValueError("name required when tables is not a dict")
            else:
                tables_to_add = {name: tables}
        else:
            tables_to_add = tables

        for df_name, df in tables_to_add.items():
            if not (isinstance(df, pd.DataFrame) or isinstance(df, pd.Series)):
                raise ValueError(
                    "table type not recognized, should be pandas DataFrame or Series."
                )
            if df_name in self._data:
                if not self.root.is_writing_mode():
                    raise IOError(f"Cannot overwrite table {df_name} in read-only mode")
                elif self.root.is_reading_mode():
                    logger.warning(f"Overwriting table: {df_name}")

            self._data[str(df_name)] = df

    def get_tables_merged(self) -> pd.DataFrame:
        """Return all tables of a model merged into one dataframe."""
        # This is mostly used for convenience and testing.
        return pd.concat(
            [df.assign(table_origin=name) for name, df in self.data.items()], axis=0
        )

    def test_equal(self, other: ModelComponent) -> tuple[bool, dict[str, str]]:
        """Test if two components are equal.

        Parameters
        ----------
        other : ModelComponent
            The component to compare against.

        Returns
        -------
        tuple[bool, dict[str, str]]
            True if the components are equal, and a dict with the associated errors per property checked.
        """
        eq, errors = super().test_equal(other)
        if not eq:
            return eq, errors
        other_tables = cast(TablesComponent, other)
        for name, df in self.data.items():
            if name not in other_tables.data:
                errors[name] = "Table not found in other component."
            elif not df.equals(other_tables.data[name]):
                errors[name] = "Table content is not equal."

        return len(errors) == 0, errors



"""A module for all of the type definitions used in HydroMT."""

from hydromt.typing.crs import CRS
from hydromt.typing.fsspec_types import FSSpecFileSystem
from hydromt.typing.metadata import SourceMetadata
from hydromt.typing.model_mode import ModelMode
from hydromt.typing.type_def import (
    Bbox,
    Crs,
    Data,
    DataType,
    ExportConfigDict,
    GeoDataframeSource,
    GeoDatasetSource,
    Geom,
    GeomBuffer,
    GpdShapeGeom,
    ModeLike,
    Number,
    Predicate,
    RasterDatasetSource,
    SourceSpecDict,
    StrPath,
    TimeRange,
    TotalBounds,
    Variables,
    XArrayDict,
    Zoom,
)

__all__ = [
    "Bbox",
    "Crs",
    "CRS",
    "StrPath",
    "ExportConfigDict",
    "FSSpecFileSystem",
    "GeoDataframeSource",
    "GeoDatasetSource",
    "ModeLike",
    "Number",
    "RasterDatasetSource",
    "SourceSpecDict",
    "TimeRange",
    "TotalBounds",
    "XArrayDict",
    "ModelMode",
    "Variables",
    "Geom",
    "GpdShapeGeom",
    "Data",
    "DataType",
    "GeomBuffer",
    "Predicate",
    "Zoom",
    "SourceMetadata",
]



import contextlib
import os


@contextlib.contextmanager
def temp_env(*remove, **update):
    """
    Temporarily updates the ``os.environ`` dictionary in-place.

    from: https://github.com/laurent-laporte-pro/stackoverflow-q2059482
    The ``os.environ`` dictionary is updated in-place so that the modification
    is sure to work in all situations.

    :param remove: Environment variables to remove.
    :param update: Dictionary of environment variables and values to add/update.
    """
    env = os.environ
    update = update or {}
    remove = remove or []

    # List of environment variables being updated or removed.
    stomped = (set(update.keys()) | set(remove)) & set(env.keys())
    # Environment variables and values to restore on exit.
    update_after = {k: env[k] for k in stomped}
    # Environment variables and values to remove on exit.
    remove_after = frozenset(k for k in update if k not in env)

    try:
        env.update(update)
        [env.pop(k, None) for k in remove]
        yield
    finally:
        env.update(update_after)
        [env.pop(k) for k in remove_after]



"""Factory function for DataSource."""

from inspect import isabstract
from typing import Any, Dict, Union

from hydromt.data_catalog.sources import (
    DataFrameSource,
    DatasetSource,
    DataSource,
    GeoDataFrameSource,
    GeoDatasetSource,
    RasterDatasetSource,
)
from hydromt.typing.type_def import DataType

# Map DataType to DataSource, need to add here when implementing a new Type
available_sources: Dict[DataType, DataSource] = {
    "DataFrame": DataFrameSource,
    "Dataset": DatasetSource,
    "RasterDataset": RasterDatasetSource,
    "GeoDataFrame": GeoDataFrameSource,
    "GeoDataset": GeoDatasetSource,
}


def create_source(data: Union[Dict[str, Any], DataSource]) -> DataSource:
    """Create a DataSource.

    Create a datasource from a dictionary, or another DataSource.
    """
    if isinstance(data, DataSource):
        if isabstract(DataSource):
            raise ValueError("DataSource is an Abstract Class")
        else:
            # Already is a subclass of DataSource
            return data

    elif isinstance(data, dict):
        if data_type := data.get("data_type", None):
            if target_source := available_sources.get(data_type):
                return target_source.model_validate(data)

            raise ValueError(f"Unknown 'data_type': '{data_type}'")
        else:
            raise ValueError("DataSource needs 'data_type'.")
    else:
        raise ValueError(f"Invalid argument for creating DataSource: {data}")



"""GeoDatasetDriver for zarr data."""

import logging
from functools import partial
from os.path import splitext
from typing import Any, Callable, ClassVar

import xarray as xr
from pydantic import Field

from hydromt._utils.unused_kwargs import _warn_on_unused_kwargs
from hydromt.data_catalog.drivers.base_driver import (
    DRIVER_OPTIONS_DESCRIPTION,
)
from hydromt.data_catalog.drivers.geodataset.geodataset_driver import (
    GeoDatasetDriver,
    GeoDatasetOptions,
)
from hydromt.error import NoDataStrategy, exec_nodata_strat
from hydromt.typing import (
    Geom,
    SourceMetadata,
    StrPath,
    TimeRange,
)
from hydromt.typing.type_def import Predicate

logger = logging.getLogger(__name__)

_ZARR_EXT = ".zarr"
_NETCDF_EXT = [".nc", ".netcdf"]


class GeoDatasetXarrayDriver(GeoDatasetDriver):
    """
    Driver for GeoDataset using the xarray library: ``geodataset_xarray``.

    Supports reading and writing zarr and netcdf files using xarray.
    zarr files will be read using `xr.open_zarr` and netcdf files using
    `xr.open_mfdataset`.
    """

    name: ClassVar[str] = "geodataset_xarray"
    supports_writing = True
    SUPPORTED_EXTENSIONS: ClassVar[set[str]] = {_ZARR_EXT, *_NETCDF_EXT}
    options: GeoDatasetOptions = Field(
        default_factory=GeoDatasetOptions, description=DRIVER_OPTIONS_DESCRIPTION
    )

    def read(
        self,
        uris: list[str],
        *,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
        kwargs_for_open: dict[str, Any] | None = None,
        mask: Geom | None = None,
        metadata: SourceMetadata | None = None,
        predicate: Predicate = "intersects",
        time_range: TimeRange | None = None,
        variables: str | list[str] | None = None,
    ) -> xr.Dataset:
        """
        Read zarr data to an xarray DataSet.

        Args:
        """
        _warn_on_unused_kwargs(
            self.__class__.__name__,
            {
                "mask": mask,
                "time_range": time_range,
                "variables": variables,
                "predicate": predicate,
            },
        )

        preprocessor = self.options.get_preprocessor()
        first_ext = splitext(uris[0])[-1]
        kwargs_for_open = kwargs_for_open or {}
        kwargs = self.options.get_kwargs() | kwargs_for_open
        if first_ext == _ZARR_EXT:
            opn: Callable = partial(xr.open_zarr, **kwargs)
            datasets = []
            for _uri in uris:
                ext = splitext(_uri)[-1]
                if ext != first_ext:
                    logger.warning(f"Reading zarr and {_uri} was not, skipping...")
                else:
                    datasets.append(
                        preprocessor(opn(_uri)) if preprocessor else opn(_uri)
                    )

            ds: xr.Dataset = xr.merge(datasets)
        elif first_ext in _NETCDF_EXT:
            filtered_uris = []
            for _uri in uris:
                ext = splitext(_uri)[-1]
                if ext != first_ext:
                    logger.warning(f"Reading netcdf and {_uri} was not, skipping...")
                else:
                    filtered_uris.append(_uri)

            ds: xr.Dataset = xr.open_mfdataset(
                filtered_uris, decode_coords="all", preprocess=preprocessor, **kwargs
            )
        else:
            raise ValueError(
                f"Unknown extention for GeoDatasetXarrayDriver: {first_ext} "
            )
        for variable in ds.data_vars:
            if ds[variable].size == 0:
                exec_nodata_strat(
                    f"No data from driver: '{self.name}' for variable: '{variable}'",
                    strategy=handle_nodata,
                )
        return ds

    def write(self, path: StrPath, ds: xr.Dataset, **kwargs) -> str:
        """
        Write the GeoDataset to a local file using zarr.

        args:
        """
        ext = splitext(path)[-1]
        if ext == _ZARR_EXT:
            ds.vector.to_zarr(path, **kwargs)
        elif ext in _NETCDF_EXT:
            ds.vector.to_netcdf(path, **kwargs)
        else:
            raise ValueError(f"Unknown extension for GeoDatasetXarrayDriver: {ext} ")

        return path



# -*- coding: utf-8 -*-
"""HydroMT VectorComponent class definition."""

import logging
from os.path import basename, dirname, isfile, join
from typing import TYPE_CHECKING, Optional, Union, cast

import geopandas as gpd
import numpy as np
import xarray as xr
from geopandas.testing import assert_geodataframe_equal
from pyproj import CRS
from shapely.geometry import box

from hydromt.gis.vector import GeoDataset
from hydromt.io.readers import open_ncs
from hydromt.io.writers import write_nc
from hydromt.model.components.base import ModelComponent
from hydromt.model.components.spatial import SpatialModelComponent
from hydromt.model.steps import hydromt_step

if TYPE_CHECKING:
    from hydromt.model.model import Model

__all__ = ["VectorComponent"]

logger = logging.getLogger(__name__)


class VectorComponent(SpatialModelComponent):
    """ModelComponent class for vector components.

    This class is used to manage vector data in a model (e.g. for polygons of a semi
    distributed model). The vector component data stored in the ``data`` property of
    this class if of the hydromt.gis.vector.GeoDataset type which is an extension of
    xarray.Dataset with a geometry coordinate.
    """

    def __init__(
        self,
        model: "Model",
        *,
        region_component: Optional[str] = None,
        region_filename: str = "vector/vector_region.geojson",
    ) -> None:
        """Initialize a vector component.

        Parameters
        ----------
        model : Model
            Parent model
        region_component : str, optional
            The name of the region component to use as reference for this component's region.
            If None, the region will be set to the bounds of the geometry of this vector component.
        region_filename : str
            The path to use for writing the region data to a file. By default "vector/vector_region.geojson".
        """
        super().__init__(
            model, region_component=region_component, region_filename=region_filename
        )
        self._data: Optional[xr.Dataset] = None

    @property
    def data(self) -> xr.Dataset:
        """Model vector (polygon) data.

        Returns xr.Dataset with a polygon geometry coordinate.
        """
        if self._data is None:
            self._initialize()
        assert self._data is not None
        return self._data

    @property
    def _region_data(self) -> Optional[gpd.GeoDataFrame]:
        if len(self.data.vector) == 0:
            return None
        gdf = self.geometry.to_frame("geometry")
        return gpd.GeoDataFrame(geometry=[box(*gdf.total_bounds)], crs=gdf.crs)

    def _initialize(self, skip_read=False) -> None:
        if self._data is None:
            self._data = xr.Dataset()
            if self.root.is_reading_mode() and not skip_read:
                self.read()

    def set(
        self,
        data: Union[xr.DataArray, xr.Dataset, np.ndarray, gpd.GeoDataFrame] = None,
        *,
        name: Optional[str] = None,
        overwrite_geom: bool = False,
    ) -> None:
        """Add data to vector.

        All layers of data must have identical spatial index.
        Only polygon geometry is supported.

        If vector already contains a geometry layer different than data,
        `overwrite_geom` can be set to True to overwrite the complete vector
        object wit data (use with caution as previous data could be lost).

        Parameters
        ----------
        data: xarray.DataArray or xarray.Dataset or np.ndarray or gpd.GeoDataFrame
            new data to add to vector
        name: str, optional
            Name of new data, this is used to overwrite the name of a DataArray
            or to select a variable from a Dataset.
        overwrite_geom: bool, optional
            If True, overwrite the complete vector object with data, by default False
        """
        self._initialize()
        assert self._data is not None
        name_required = isinstance(data, np.ndarray) or (
            isinstance(data, xr.DataArray) and data.name is None
        )
        if name is None and name_required:
            raise ValueError(f"Unable to set {type(data).__name__} data without a name")

        # check the type of data
        if isinstance(data, np.ndarray) and "geometry" in self._data:
            index_dim = self.index_dim
            index = self._data[index_dim]
            if data.size != index.size:
                if data.ndim == 1:
                    raise ValueError("Size of data and number of vector do not match")
                else:
                    raise ValueError(
                        "set_vector with np.ndarray is only supported if data is 1D"
                    )
            data = xr.DataArray(dims=[index_dim], data=data)
        if isinstance(data, xr.DataArray):
            if name is not None:  # rename
                data.name = name
            data = data.to_dataset()
        elif isinstance(data, gpd.GeoDataFrame):
            data = GeoDataset.from_gdf(data, keep_cols=True, cols_as_data_vars=True)
        elif not isinstance(data, xr.Dataset):
            raise ValueError(f"cannot set data of type {type(data).__name__}")

        # Add to vector
        # 1. self.vector does not have a geometry yet or overwrite_geom
        # then use data directly
        if self.geometry is None or overwrite_geom:
            if data.vector.geometry is None:
                raise ValueError("Cannot instantiate vector without geometry in data")
            else:
                if overwrite_geom:
                    logger.warning("Overwriting vector object with data")
                self._data = data
        # 2. self.vector has a geometry
        else:
            # data has a geometry - check if it is the same as self.vector
            if data.vector.geometry is not None and not np.all(
                data.vector.geometry.geom_equals_exact(self.geometry, tolerance=0.0001)
            ):
                raise ValueError("Geometry of data and vector do not match")
            # add data (with check on index)
            for dvar in data.data_vars:
                if dvar in self._data:
                    logger.warning(f"Replacing vector variable: {dvar}")
                # check on index coordinate before merging
                dims = data[dvar].dims
                if np.array_equal(
                    data[dims[0]].values, self._data[self.index_dim].values
                ):
                    self._data[dvar] = data[dvar]
                else:
                    raise ValueError(
                        f"Index coordinate of data variable {dvar} "
                        "does not match vector index coordinate"
                    )

    @hydromt_step
    def read(
        self,
        *,
        filename: Optional[str] = "vector/vector.nc",
        geometry_filename: Optional[str] = "vector/vector.geojson",
        **kwargs,
    ) -> None:
        """Read model vector from combined netcdf and geojson file.

        Files are read at <root>/<filename> and geojson file at <root>/<geometry_filename>.

        Three options are possible:
            * The netcdf file contains the attribute data and the geojson file the
                geometry vector data.
            * The netcdf file contains both the attribute and the geometry data.
                (geometry_filename is ignored)
            * The geojson file contains both the attribute and the geometry data.
                (filename is ignored)

        Key-word arguments are passed to :py:meth:`~hydromt.model.Model.open_nc`

        Parameters
        ----------
        filename : str, optional
            netcdf filename relative to model root,
            by default 'vector/vector.nc'
        geometry_filename : str, optional
            geojson filename relative to model root,
            by default 'vector/vector.geojson'
        kwargs:
            Additional keyword arguments that are passed to the `open_nc`
            function.
        """
        self.root._assert_read_mode()
        self._initialize(skip_read=True)

        if filename is None and geometry_filename is None:
            raise ValueError(
                "Both filename and geometry_filename are None, no source file given."
            )

        if filename is not None:
            # Disable lazy loading of data
            # to avoid issues with reading object dtype data
            kwargs.setdefault("chunks", None)
            files = open_ncs(filename, root=self.root.path, **kwargs).values()
            self._open_datasets.extend(files)
            ds = xr.merge(files)
            # check if ds is empty (default filename has a value)
            if len(ds.sizes) == 0:
                filename = None
        if geometry_filename is not None and isfile(
            join(self.root.path, geometry_filename)
        ):
            gdf = gpd.read_file(join(self.root.path, geometry_filename))
            # geom + netcdf data
            if filename is not None:
                ds = GeoDataset.from_gdf(gdf, data_vars=ds)
            # geom only
            else:
                ds = GeoDataset.from_gdf(gdf, keep_cols=True, cols_as_data_vars=True)
        # netcdf only
        elif filename is not None:
            ds = GeoDataset.from_netcdf(ds)
        else:
            logger.info("No vector data found, skip reading.")
            return

        self.set(data=ds)

    @hydromt_step
    def write(
        self,
        filename: Optional[str] = "vector/vector.nc",
        geometry_filename: Optional[str] = "vector/vector.geojson",
        *,
        ogr_compliant: bool = False,
        **kwargs,
    ) -> None:
        """Write model vector to combined netcdf and geojson files.

        Files are written at <root>/<filename> and at <root>/<geometry_filename> respectively.

        Three options are possible:

            * The netcdf file contains the attribute data and the geojson file the
                geometry vector data. Key-word arguments are passed to
                :py:meth:`~hydromt.model.Model.write_nc`

            * The netcdf file contains both the attribute and the geometry data
                (geometry_filename set to None). Key-word arguments are passed to
                :py:meth:`~hydromt.model.Model.write_nc`

            * The geojson file contains both the attribute and the geometry data
                (filename set to None). This option is possible only if all data variables
                are 1D. Else the data will be written to netcdf. Key-word arguments are
                passed to :py:meth:`~hydromt.vector.GeoDataset.to_gdf`

        Parameters
        ----------
        filename : str, optional
            netcdf filename relative to model root,
            by default 'vector/vector.nc'
        geometry_filename : str, optional
            geojson filename relative to model root,
            by default 'vector/vector.geojson'
        ogr_compliant : bool
            If filename only, write the netCDF4 file in an ogr compliant format
            This makes it readable as a vector file in e.g. QGIS
            see :py:meth:`~hydromt.vector.GeoBase.ogr_compliant` for more details.
        **kwargs:
            Additional keyword arguments that are passed to the `write_nc`
            function.
        """
        self.root._assert_write_mode()

        ds = self.data
        if len(ds) == 0:
            logger.info(
                f"{self.model.name}.{self.name_in_model}: No vector data found, skip writing."
            )
            return

        if filename is None and geometry_filename is None:
            raise ValueError(
                "Both filename and geometry_filename are None, no destination file given. Please provide either filename or geometry_filename."
            )

        # If filename is None check if vector contains only 1D data
        if filename is None:
            assert geometry_filename is not None
            # Check if 1D data only is present
            snames = ["y_name", "x_name", "index_dim", "geom_name"]
            sdims = [ds.vector.attrs.get(n) for n in snames if n in ds.vector.attrs]
            if "spatial_ref" in ds:
                sdims.append("spatial_ref")
            for name in set(ds.vector._all_names) - set(sdims):
                dims = ds[name].dims
                # check 1D variables with matching index_dim
                if len(dims) > 1 or dims[0] != ds.vector.index_dim:
                    filename = join(
                        dirname(join(self.root.path, geometry_filename)),
                        f"{basename(geometry_filename).split('.')[0]}.nc",
                    )
                    logger.warning(
                        "2D data found in vector,"
                        f"will write data to {filename} instead."
                    )
                    break

        # write to netcdf only
        if geometry_filename is None:
            assert filename is not None
            # cannot call directly ds.vector.to_netcdf
            # because of possible PermissionError
            if ogr_compliant:
                ds = ds.vector.ogr_compliant()
            else:
                ds = ds.vector.update_geometry(geom_format="wkt", geom_name="ogc_wkt")
            # write_nc requires dict - use dummy key
            geojson_path = self.root.path / filename
            geojson_path.parent.mkdir(parents=True, exist_ok=True)
            logger.info(
                f"{self.model.name}.{self.name_in_model}: Writing vector to {geojson_path}."
            )
            close_handle = write_nc(
                ds,
                file_path=geojson_path,
                engine="netcdf4",
                force_overwrite=self.root.mode.is_override_mode(),
                **kwargs,
            )
            if close_handle is not None:
                self._deferred_file_close_handles.append(close_handle)
        # write to geojson only
        elif filename is None:
            geojson_path = self.root.path / geometry_filename
            geojson_path.parent.mkdir(parents=True, exist_ok=True)
            logger.info(
                f"{self.model.name}.{self.name_in_model}: Writing vector to {geojson_path}."
            )
            gdf = ds.vector.to_gdf(**kwargs)
            gdf.to_file(geojson_path)
        # write data to netcdf and geometry to geojson
        else:
            geojson_path = self.root.path / geometry_filename
            geojson_path.parent.mkdir(parents=True, exist_ok=True)
            nc_path = self.root.path / filename
            nc_path.parent.mkdir(parents=True, exist_ok=True)
            logger.info(
                f"{self.model.name}.{self.name_in_model}: Writing geometry to {geojson_path} and data to {nc_path}."
            )

            gdf = ds.vector.geometry.to_frame("geometry")
            gdf.to_file(geojson_path)

            close_handle = write_nc(
                ds.drop_vars("geometry"),
                file_path=nc_path,
                **kwargs,
            )
            if close_handle is not None:
                self._deferred_file_close_handles.append(close_handle)

    # Other vector properties
    @property
    def geometry(self) -> gpd.GeoSeries:
        """Returns the geometry of the model vector as gpd.GeoSeries."""
        # check if vector is empty
        if len(self.data.sizes) == 0:
            return None
        else:
            return self.data.vector.geometry

    @property
    def index_dim(self) -> str:
        """Returns the index dimension of the vector."""
        return self.data.vector.index_dim

    @property
    def crs(self) -> Optional[CRS]:
        """Returns coordinate reference system embedded in the vector."""
        if self.data.vector.crs is not None:
            return self.data.vector.crs
        logger.warning("No CRS found in vector data.")
        return None

    def test_equal(self, other: ModelComponent) -> tuple[bool, dict[str, str]]:
        """Test if two components are equal.

        Parameters
        ----------
        other : ModelComponent
            The component to compare against.

        Returns
        -------
        tuple[bool, dict[str, str]]
            True if the components are equal, and a dict with the associated errors per property checked.
        """
        eq, errors = super().test_equal(other)
        if not eq:
            return eq, errors
        other_vector = cast(VectorComponent, other)

        geods = self.data
        other_geods = other_vector.data
        gdf = self.geometry.to_frame("geometry")
        gdf_other = other_vector.geometry.to_frame("geometry")
        try:
            assert_geodataframe_equal(
                gdf, gdf_other, check_like=True, check_less_precise=True
            )
        except AssertionError as e:
            errors["geometry"] = str(e)
        drop_vars = (
            [geods.vector.x_name, geods.vector.y_name]
            if geods.vector.geom_format == "xy"
            else [geods.vector.geom_name]
        )
        ds = geods.drop_vars(drop_vars)
        ds_other = other_geods.drop_vars(drop_vars)
        try:
            xr.testing.assert_allclose(ds, ds_other)
        except AssertionError as e:
            errors["data"] = str(e)

        return len(errors) == 0, errors



"""Caching mechanisms used in HydroMT."""

import logging
import os
import xml.etree.ElementTree as ET
from ast import literal_eval
from os.path import basename, dirname, isdir, isfile, join
from pathlib import Path
from typing import List, Optional, cast

import geopandas as gpd
import numpy as np
from affine import Affine
from fsspec import AbstractFileSystem, url_to_fs
from pyproj import CRS

from hydromt._compat import HAS_GDAL
from hydromt._utils.uris import _strip_scheme, _strip_vsi
from hydromt.config import SETTINGS

if HAS_GDAL:
    from osgeo import gdal

    gdal.UseExceptions()


logger = logging.getLogger(__name__)

__all__ = ["_copy_to_local", "_cache_vrt_tiles"]


def _copy_to_local(src: str, dst: Path, fs: Optional[AbstractFileSystem] = None):
    """Copy files from source uri to local file.

    Parameters
    ----------
    src : str
        Source URI
    dst : Path
        Destination path
    fs : Optional[AbstractFileSystem], optional
        Fsspec filesystem. Will be inferred from src if not supplied.
    """
    if fs is None:
        fs = cast(AbstractFileSystem, url_to_fs(src)[0])
    if not isdir(dirname(dst)):
        os.makedirs(dirname(dst), exist_ok=True)

    fs.get(src, str(dst))


def _overlaps(source: ET.Element, affine: Affine, bbox: List[float]) -> bool:
    """Check whether source overlaps with bbox.

    Parameters
    ----------
    source : ET.Element
        Source element in .vrt
    affine : Affine
        Tile affine
    bbox : List[float]
        Requested bbox

    Returns
    -------
    bool
        Whether the tile overlaps with the requested bbox.

    Raises
    ------
    ValueError
        If tile bbox of tile not available as 'DstRect' sub element.
    """
    names = ["xOff", "yOff", "xSize", "ySize"]
    dest_rectangle: Optional[ET.Element] = source.find("DstRect")
    if dest_rectangle is None:
        raise ValueError("Tile metadata missing.")
    x0, y0, dx, dy = [float(dest_rectangle.get(name)) for name in names]

    xs, ys = affine * (np.array([x0, x0 + dx]), np.array([y0, y0 + dy]))

    return (
        max(xs) > bbox[0]
        and max(ys) > bbox[1]
        and min(xs) < bbox[2]
        and min(ys) < bbox[3]
    )


def _cache_vrt_tiles(
    vrt_uri: str,
    fs: Optional[AbstractFileSystem] = None,
    geom: Optional[gpd.GeoSeries] = None,
    cache_dir: str | Path = SETTINGS.cache_root,
) -> Path:
    """Cache vrt tiles that intersect with geom.

    Note that the vrt file must contain relative tile paths.

    Parameters
    ----------
    vrt_uri: str
        Path to source vrt
    fs : Optional[AbstractFileSystem], optional
        Fsspec filesystem. Will be inferred from src if not supplied.
    geom: geopandas.GeoSeries, optional
        Geometry to intersect tiles with
    cache_dir: str, Path
        Path of the root folder where

    Returns
    -------
    vrt_destination_path : Path
        Path to cached vrt
    """
    if not HAS_GDAL:
        raise ImportError("Can't cache vrt's without GDAL installed.")

    # Get the filesystem type
    if fs is None:
        fs: AbstractFileSystem = url_to_fs(vrt_uri)[0]

    # cache vrt file
    vrt_root: str = fs._parent(vrt_uri)

    # Strip scheme from base uri and get path to directory with vrt file
    scheme, stripped = _strip_scheme(vrt_uri)

    # read vrt file
    root = ET.fromstring(fs.read_text(vrt_uri))

    # get vrt transform and crs
    geotransform_el: Optional[ET.Element] = root.find("GeoTransform")
    if geotransform_el is None:
        raise ValueError(f"No GeoTransform found in: {vrt_uri}")
    transform = Affine.from_gdal(*literal_eval(geotransform_el.text.strip()))

    srs_el: Optional[ET.Element] = root.find("SRS")
    if srs_el is None:
        raise ValueError(f"No SRS info found at: {vrt_uri}")

    crs: CRS = CRS.from_string(srs_el.text)
    # get geometry bbox in vrt crs
    if geom is not None:
        if crs != geom.crs:
            geom = geom.to_crs(crs)
        bbox = geom.total_bounds
    # support multiple type of sources in vrt
    band_name: Optional[ET.Element] = root.find("VRTRasterBand")
    if band_name is None:
        raise ValueError(f"Could not find VRTRasterBand in: {vrt_uri}")
    try:
        source_name: str = next(
            filter(lambda k: k.tag.endswith("Source"), band_name.iter())
        ).tag
    except StopIteration:
        raise ValueError(f"No Source information found at: {vrt_uri}")

    # loop through files in VRT and check if in bbox
    paths = []
    for source in band_name.findall(source_name):
        if geom is None or _overlaps(source, transform, bbox):
            vrt_ref_el: Optional[ET.Element] = source.find("SourceFilename")
            if vrt_ref_el is None:
                raise ValueError(f"Could not find Source File in vrt: {vrt_uri}.")
            vrt_ref: str = vrt_ref_el.text.replace("\\", os.sep)  # dewindowsify
            vrt_relative: int = int(vrt_ref_el.get("relativeToVRT"))  # 0 or 1

            if not vrt_relative:
                if vrt_ref.startswith("/vsi"):
                    # virtual file system https://gdal.org/user/virtual_file_systems.html
                    # also starts with /vsi on windows

                    # get base path vrs, can safely strip /vsi<type>/
                    _, vrt_ref = _strip_vsi(vrt_ref)

                if scheme is None:
                    # local path
                    # Get relative uri that the vrt points to
                    relative_uri_ref: Path = Path(
                        os.path.relpath(vrt_ref, dirname(stripped))
                    )
                    # Set download uri to match vrt_uri
                    src_uri: str = vrt_ref
                else:
                    # matching fs sep
                    sep = fs.sep
                    vrt_directory: str = sep.join(stripped.split(sep)[:-1])
                    # Get relative uri that the vrt points to
                    relative_uri_ref: str = vrt_ref.lstrip(vrt_directory)
                    # Set download uri to match vrt_uri
                    src_uri: str = scheme + vrt_ref

                # Set download dest to cache_dir and refer to it in the vrt
                dst: Path = cache_dir / relative_uri_ref
                vrt_ref_el.text = relative_uri_ref
            else:
                if scheme is None:
                    # local uri
                    src_uri: str = join(vrt_root, vrt_ref)
                else:
                    # fsspec uri
                    src_uri: str = fs.sep.join((vrt_root), vrt_ref)

                dst: Path = cache_dir / vrt_ref
            if not isfile(dst):  # Skip cached files
                paths.append((src_uri, dst))

    # Define the output vrt
    vrt_destination_path = Path(cache_dir, basename(vrt_uri))

    # Copy the new files
    logger.info(f"Downloading {len(paths)} tiles to {cache_dir}")
    for src, dst in paths:
        _copy_to_local(src, dst, fs)

    # Check for the newly cached files
    new = [item[1] for item in paths]
    if len(new) == 0:
        if vrt_destination_path.is_file():
            return vrt_destination_path
        else:
            return Path(vrt_uri)  # The original and try to work with that

    # Second check for existing vrt and add the files to list to build the new vrt
    if vrt_destination_path.is_file():
        with open(vrt_destination_path, "r") as f:
            root = ET.fromstring(f.read())
        cur = [
            Path(vrt_destination_path.parent, item.find("SourceFilename").text)
            for item in root.find("VRTRasterBand").findall(source_name)
        ]
        # This shouldnt be possible, but just to be sure.
        cur = [item for item in cur if item not in new]
        # Add the current files in the vrt to the list and unlink the current vrt
        new = new + cur
        os.unlink(vrt_destination_path)

    # Build the vrt with gdal
    out_ds = gdal.BuildVRT(
        destName=vrt_destination_path.as_posix(),
        srcDSOrSrcDSTab=new,
    )

    # Close and dereference the gdal dataset
    out_ds.Close()
    out_ds = None

    return vrt_destination_path



"""Runtime Settings for HydroMT."""

from pathlib import Path
from typing import Annotated, Any

from pydantic import Field, ValidationInfo, ValidatorFunctionWrapHandler, WrapValidator
from pydantic_settings import BaseSettings


def _validate_path(
    path: Any, handler: ValidatorFunctionWrapHandler, info: ValidationInfo
):
    if isinstance(path, str):
        path = Path(path)
    return handler(path, info)


Pathdantic = Annotated[Path, WrapValidator(_validate_path)]


class Settings(BaseSettings):
    """Runtime Settings for HydroMT."""

    cache_root: Pathdantic = Field(default=Path.home() / ".hydromt")


SETTINGS = Settings()



"""Driver using rasterio for RasterDataset."""

import logging
from pathlib import Path
from typing import Any, ClassVar

import numpy as np
import rasterio
import rasterio.errors
import xarray as xr
from pydantic import Field, field_serializer, model_validator
from pyproj import CRS

from hydromt._utils.caching import _cache_vrt_tiles
from hydromt._utils.temp_env import temp_env
from hydromt._utils.uris import _strip_scheme
from hydromt.config import SETTINGS
from hydromt.data_catalog.drivers.base_driver import DriverOptions
from hydromt.data_catalog.drivers.raster.raster_dataset_driver import (
    RasterDatasetDriver,
)
from hydromt.error import NoDataStrategy, exec_nodata_strat
from hydromt.gis.gis_utils import zoom_to_overview_level
from hydromt.io.readers import open_mfraster
from hydromt.typing import (
    Geom,
    SourceMetadata,
    StrPath,
    TimeRange,
    Variables,
    Zoom,
)

logger = logging.getLogger(__name__)


class RasterioOptions(DriverOptions):
    """Options for RasterioDriver."""

    KWARGS_FOR_OPEN: ClassVar[set[str]] = {"mosaic_kwargs"}

    mosaic: bool = Field(
        default=False,
        description="If True and multiple uris are given, will mosaic the datasets together using `rasterio.merge.merge`. Default is False.",
    )

    mosaic_kwargs: dict[str, Any] = Field(
        default={},
        description="Additional keyword arguments to pass to `rasterio.merge.merge`.",
    )

    cache: bool = Field(
        default=False,
        description="If True and reading from VRT files, will cache the tiles locally to speed up reading. Default is False.",
    )

    cache_root: str = Field(
        default=str(SETTINGS.cache_root),
        description="Root directory for caching. Default is taken from `hydromt.config.SETTINGS.cache_root`.",
    )

    cache_dir: str | None = Field(
        default=None,
        description="Subdirectory for caching. Default is the stem of the first uri without extension.",
    )

    def get_cache_path(self, uris: list[str]) -> Path:
        """Get the cache path based on the options and uris."""
        if self.cache_dir is not None:
            cache_dir = Path(self.cache_root) / self.cache_dir
        else:
            # default to first uri without extension
            cache_dir = Path(self.cache_root) / Path(_strip_scheme(uris[0])[1]).stem
        return cache_dir

    @model_validator(mode="after")
    def _convert_path_to_str(self):
        """Convert Path to str for pydantic compatibility."""
        if isinstance(self.cache_root, Path):
            self.cache_root = self.cache_root.as_posix()
        if isinstance(self.cache_dir, Path):
            self.cache_dir = self.cache_dir.as_posix()
        return self

    @field_serializer("cache_root", "cache_dir")
    def serialize_paths(self, value: Path) -> str | None:
        """Serialize Path to str for pydantic compatibility."""
        if value is None:
            return None
        return Path(value).as_posix()


class RasterioDriver(RasterDatasetDriver):
    """
    Driver for RasterDataset using the rasterio library: ``rasterio``.

    Supports reading and writing raster files using rasterio.
    """

    name = "rasterio"
    SUPPORTED_EXTENSIONS: ClassVar[set[str]] = {
        "." + extension
        for extension in rasterio.drivers.raster_driver_extensions()
        if extension != "nc"
    }  # Exclude netcdf as a supported file type
    options: RasterioOptions = Field(default_factory=RasterioOptions)

    def read(
        self,
        uris: list[str],
        *,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
        kwargs_for_open: dict[str, Any] | None = None,
        mask: Geom | None = None,
        variables: Variables | None = None,
        time_range: TimeRange | None = None,
        zoom: Zoom | None = None,
        chunks: dict[str, Any] | None = None,
        metadata: SourceMetadata | None = None,
    ) -> xr.Dataset:
        """Read data using rasterio."""
        if metadata is None:
            metadata = SourceMetadata()

        # Caching portion, only when the flag is True and the file format is vrt
        if all([uri.endswith(".vrt") for uri in uris]) and self.options.cache:
            cache_dir: Path = self.options.get_cache_path(uris)
            uris_cached = []
            for uri in uris:
                cached_uri: str = _cache_vrt_tiles(
                    uri, geom=mask, fs=self.filesystem.get_fs(), cache_dir=cache_dir
                )
                uris_cached.append(cached_uri)
            uris = uris_cached

        if mask is not None:
            self.options.mosaic_kwargs.update({"mask": mask})

        kwargs_for_open = kwargs_for_open or {}
        if np.issubdtype(type(metadata.nodata), np.number):
            kwargs_for_open.update({"nodata": metadata.nodata})

        # Fix overview level
        if zoom:
            try:
                zls_dict: dict[int, float] = metadata.zls_dict
                crs: CRS | None = metadata.crs
            except AttributeError:  # pydantic extra=allow on SourceMetadata
                zls_dict, crs = self._get_zoom_levels_and_crs(uris[0])

            overview_level: int | None = zoom_to_overview_level(
                zoom, mask, zls_dict, crs
            )
            if overview_level:
                # NOTE: overview levels start at zoom_level 1, see _get_zoom_levels_and_crs
                kwargs_for_open.update(overview_level=overview_level - 1)

        if chunks is not None:
            kwargs_for_open.update({"chunks": chunks})

        kwargs = self.options.get_kwargs() | kwargs_for_open
        mosaic: bool = self.options.mosaic and len(uris) > 1

        # If the metadata resolver has already resolved the overview level,
        # trying to open zoom levels here will result in an error.
        # Better would be to separate uriresolver and driver: https://github.com/Deltares/hydromt/issues/1023
        # Then we can implement looking for a overview level in the driver.
        def _open() -> xr.DataArray | xr.Dataset:
            try:
                return open_mfraster(
                    uris,
                    mosaic=mosaic,
                    **kwargs,
                )
            except rasterio.errors.RasterioIOError as e:
                if "Cannot open overview level" in str(e):
                    kwargs.pop("overview_level", None)
                    return open_mfraster(
                        uris,
                        mosaic=mosaic,
                        **kwargs,
                    )
                else:
                    raise

        # rasterio uses specific environment variable for s3 access.
        try:
            anon: str = self.filesystem.get_fs().anon
        except AttributeError:
            anon: str = ""

        if anon:
            with temp_env(**{"AWS_NO_SIGN_REQUEST": "true"}):
                ds = _open()
        else:
            ds = _open()

        # Mosaic's can mess up the chunking, which can error during writing
        # Or maybe setting
        chunks = kwargs.get("chunks", None)
        if chunks is not None:
            ds = ds.chunk(chunks=chunks)

        # rename ds with single band if single variable is requested
        if variables is not None and len(variables) == 1 and len(ds.data_vars) == 1:
            ds = ds.rename({list(ds.data_vars.keys())[0]: list(variables)[0]})

        for variable in ds.data_vars:
            if ds[variable].size == 0:
                exec_nodata_strat(
                    f"No data from driver: '{self.name}' for variable: '{variable}'",
                    strategy=handle_nodata,
                )
        return ds

    def write(self, path: StrPath, ds: xr.Dataset, **kwargs) -> str:
        """Write out a RasterDataset using rasterio."""
        raise NotImplementedError()

    @staticmethod
    def _get_zoom_levels_and_crs(uri: str) -> tuple[dict[int, float], int]:
        """Get zoom levels and crs from adapter or detect from tif file if missing."""
        zoom_levels = {}
        crs = None
        try:
            with rasterio.open(uri) as src:
                res = abs(src.res[0])
                crs = src.crs
                overviews = [src.overviews(i) for i in src.indexes]
                if len(overviews[0]) > 0:  # check overviews for band 0
                    # check if identical
                    if not all([o == overviews[0] for o in overviews]):
                        raise ValueError("Overviews are not identical across bands")
                    # dict with overview level and corresponding resolution
                    zls = [1] + overviews[0]
                    zoom_levels = {i: res * zl for i, zl in enumerate(zls)}
        except rasterio.RasterioIOError as e:
            logger.warning(f"IO error while detecting zoom levels: {e}")
        return zoom_levels, crs



import logging
from typing import TYPE_CHECKING, Dict, Optional, cast

import numpy as np
import pandas as pd
import xarray as xr

from hydromt._utils.nodata import _has_no_data
from hydromt.typing.metadata import SourceMetadata
from hydromt.typing.type_def import TimeRange, Variables

if TYPE_CHECKING:
    from pandas._libs.tslibs.timedeltas import TimeDeltaUnitChoices


logger = logging.getLogger(__name__)

__all__ = [
    "_set_metadata",
    "_shift_dataset_time",
    "_slice_temporal_dimension",
    "_rename_vars",
]


def _shift_dataset_time(
    dt: int,
    ds: Optional[xr.Dataset],
    time_unit: "TimeDeltaUnitChoices" = "s",
) -> Optional[xr.Dataset]:
    """Shifts time of a xarray dataset.

    Parameters
    ----------
    dt : int
        time delta to shift the time of the dataset
    ds : xr.Dataset
        xarray dataset
    logger : logging.Logger
        logger

    Returns
    -------
    xr.Dataset
        time shifted dataset
    """
    if ds is None:
        return None

    if (
        dt != 0
        and "time" in ds.dims
        and ds["time"].size > 1
        and np.issubdtype(ds["time"].dtype, np.datetime64)
    ):
        logger.debug(f"Shifting time labels with {dt} {time_unit}.")
        ds["time"] = ds["time"] + pd.to_timedelta(dt, unit=time_unit)
    elif dt != 0:
        logger.warning("Time shift not applied, time dimension not found.")
    return ds


def _slice_temporal_dimension(
    ds: Optional[xr.Dataset], time_range: TimeRange
) -> Optional[xr.Dataset]:
    if ds is None:
        return None
    else:
        if (
            "time" in ds.dims
            and ds["time"].size > 1
            and np.issubdtype(ds["time"].dtype, np.datetime64)
        ):
            logger.debug(f"Slicing time dim {time_range}")
            ds = ds.sel(time=slice(time_range.start, time_range.end))
        if _has_no_data(ds):
            return None
        else:
            return ds


def _set_metadata(
    ds: Optional[xr.Dataset], metadata: "SourceMetadata"
) -> Optional[xr.Dataset]:
    if ds is None:
        return None
    elif metadata.attrs:
        if isinstance(ds, xr.DataArray):
            name = cast(str, ds.name)
            ds.attrs.update(metadata.attrs[name])
        else:
            for k in metadata.attrs:
                ds[k].attrs.update(metadata.attrs[k])

    ds.attrs.update(metadata.model_dump(exclude_unset=True, exclude={"attrs"}))
    return ds


def _rename_vars(
    ds: Optional[xr.Dataset], rename: Dict[str, str]
) -> Optional[xr.Dataset]:
    if ds is None:
        return None
    rm = {k: v for k, v in rename.items() if k in ds}
    ds = ds.rename(rm)
    return ds


def _single_var_as_array(
    maybe_ds: Optional[xr.Dataset],
    single_var_as_array: bool,
    variable_name: Optional[Variables] = None,
) -> Optional[xr.Dataset]:
    if isinstance(maybe_ds, xr.DataArray):
        return maybe_ds
    if maybe_ds is None:
        return None
    else:
        ds = maybe_ds
    # return data array if single variable dataset
    dvars = list(ds.data_vars.keys())
    if single_var_as_array and len(dvars) == 1:
        da = ds[dvars[0]]
        if isinstance(variable_name, list) and len(variable_name) == 1:
            da.name = variable_name[0]
        elif isinstance(variable_name, str):
            da.name = variable_name
        return da
    else:
        return ds



"""Driver for reading in GeoDataFrames from tabular formats."""

import logging
from typing import Any, ClassVar

import geopandas as gpd
from pydantic import Field

from hydromt.data_catalog.drivers.base_driver import (
    DRIVER_OPTIONS_DESCRIPTION,
    DriverOptions,
)
from hydromt.data_catalog.drivers.geodataframe.geodataframe_driver import (
    GeoDataFrameDriver,
)
from hydromt.error import NoDataStrategy, exec_nodata_strat
from hydromt.io.readers import open_vector_from_table
from hydromt.typing.metadata import SourceMetadata

logger = logging.getLogger(__name__)

# possible labels for the x and y dimensions
X_DIM_LABELS = ("x", "longitude", "lon", "long")
Y_DIM_LABELS = ("y", "latitude", "lat")


class GeoDataFrameTableOptions(DriverOptions):
    """Options for the GeoDataFrameTableDriver."""

    x_dim: str | None = None
    y_dim: str | None = None


class GeoDataFrameTableDriver(GeoDataFrameDriver):
    """
    Driver for GeoDataFrame from tabular formats: ``geodataframe_table``.

    Supports reading point geometries from csv, excel (xls, xlsx), and parquet files
    using a combination of the pandas and geopandas libraries.

    """

    name: ClassVar[str] = "geodataframe_table"
    SUPPORTED_EXTENSIONS: ClassVar[set[str]] = {".csv", ".xlsx", ".xls", ".parquet"}

    options: GeoDataFrameTableOptions = Field(
        default_factory=GeoDataFrameTableOptions, description=DRIVER_OPTIONS_DESCRIPTION
    )

    def read(
        self,
        uris: list[str],
        *,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
        kwargs_for_open: dict[str, Any] | None = None,
        metadata: SourceMetadata | None = None,
        mask: Any = None,
        predicate: str = "intersects",
        variables: str | list[str] | None = None,
    ) -> gpd.GeoDataFrame:
        """Read tabular data using a combination of the pandas and geopandas libraries."""
        if not metadata:
            metadata = SourceMetadata()
        if len(uris) > 1:
            raise ValueError(
                "DataFrame: Reading multiple files with the "
                f"{self.__class__.__name__} driver is not supported."
            )

        _uri: str = uris[0]
        kwargs_for_open = kwargs_for_open or {}
        kwargs = self.options.get_kwargs() | kwargs_for_open
        gdf = open_vector_from_table(
            path=_uri,
            x_dim=self.options.x_dim,
            y_dim=self.options.y_dim,
            crs=metadata.crs,
            **kwargs,
        )
        if gdf.index.size == 0:
            exec_nodata_strat(
                f"No data from driver {self}'.",
                strategy=handle_nodata,
            )
        return gdf



"""Driver for handling IO of GeoDatasets."""

import logging
from abc import ABC, abstractmethod
from typing import Any, Callable

import xarray as xr
from pydantic import Field

from hydromt.data_catalog.drivers.base_driver import (
    DRIVER_OPTIONS_DESCRIPTION,
    BaseDriver,
    DriverOptions,
)
from hydromt.data_catalog.drivers.preprocessing import get_preprocessor
from hydromt.error import NoDataStrategy
from hydromt.typing import Geom, Predicate, SourceMetadata, StrPath, TimeRange

logger = logging.getLogger(__name__)


class GeoDatasetOptions(DriverOptions):
    """Options for GeoDatasetVectorDriver."""

    preprocess: str | None = None
    """Name of preprocessor to apply on geodataset after reading. Available preprocessors include: round_latlon, to_datetimeindex, remove_duplicates, harmonise_dims. See their docstrings for details."""

    def get_preprocessor(self) -> Callable | None:
        """Get the preprocessor function."""
        if self.preprocess is None:
            return None
        return get_preprocessor(self.preprocess)


class GeoDatasetDriver(BaseDriver, ABC):
    """Abstract Driver to read GeoDatasets."""

    options: GeoDatasetOptions = Field(
        default_factory=GeoDatasetOptions, description=DRIVER_OPTIONS_DESCRIPTION
    )

    @abstractmethod
    def read(
        self,
        uris: list[str],
        *,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
        kwargs_for_open: dict[str, Any] | None = None,
        mask: Geom | None = None,
        predicate: Predicate = "intersects",
        variables: list[str] | None = None,
        time_range: TimeRange | None = None,
        metadata: SourceMetadata | None = None,
    ) -> xr.Dataset | None:
        """
        Read in any compatible data source to an xarray Dataset.

        args:
        """
        ...

    def write(
        self,
        path: StrPath,
        ds: xr.Dataset,
        **kwargs,
    ) -> str:
        """
        Write out a GeoDataset to file.

        Not all drivers should have a write function, so this method is not
        abstract.

        args:
        """
        raise NotImplementedError(
            f"Writing using driver '{self.name}' is not supported."
        )



"""A component to write configuration files for model simulations/kernels."""

import logging
from os.path import isfile
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple, Union, cast

from hydromt._utils.path import _make_config_paths_relative
from hydromt.io.readers import read_toml, read_yaml
from hydromt.io.writers import write_toml, write_yaml
from hydromt.model.components.base import ModelComponent
from hydromt.model.steps import hydromt_step

if TYPE_CHECKING:
    from hydromt.model import Model

logger = logging.getLogger(__name__)


class ConfigComponent(ModelComponent):
    """
    A component to manage configuration files for model simulations/settings.

    ``ConfigComponent`` data is stored as a dictionary and can be written to a file
    in yaml or toml format. The component can be used to store model settings
    and parameters that are used in the model simulations or in the model
    settings.
    """

    def __init__(
        self,
        model: "Model",
        *,
        filename: str = "config.yaml",
        default_template_filename: Optional[str] = None,
    ):
        """Initialize a ConfigComponent.

        Parameters
        ----------
        model: Model
            HydroMT model instance
        filename: str
            A path relative to the root where the configuration file will
            be read and written if user does not provide a path themselves.
            By default 'config.yml'
        default_template_filename: Optional[Path]
            A path to a template file that will be used as default in the ``create``
            method to initialize the configuration file if the user does not provide
            their own template file. This can be used by model plugins to provide a
            default configuration template. By default None.
        """
        self._data: Optional[Dict[str, Any]] = None
        self._filename: str = filename
        self._default_template_filename: Optional[str] = default_template_filename

        super().__init__(model=model)

    @property
    def data(self) -> Dict[str, Any]:
        """Model config values."""
        if self._data is None:
            self._initialize()
        if self._data is None:
            raise RuntimeError("Could not load data for model config component")
        else:
            return self._data

    def _initialize(self, skip_read=False) -> None:
        """Initialize the model config."""
        if self._data is None:
            self._data = {}
            if self.root.is_reading_mode() and not skip_read:
                self.read()

    @hydromt_step
    def write(self, file_path: Optional[str] = None) -> None:
        """Write model config at <root>/{path}.

        Parameters
        ----------
        filename: str, optional
            The name of the file to write the config to. If not provided, the default
            filename will be used.
            Can be a relative path.
        """
        self.root._assert_write_mode()

        if not self.data:
            logger.info(
                f"{self.model.name}.{self.name_in_model}: No config data found, skip writing."
            )
            return

        file_path = file_path or self._filename
        file_path = self.root.path / file_path
        file_path.parent.mkdir(parents=True, exist_ok=True)
        logger.info(
            f"{self.model.name}.{self.name_in_model}: Writing model config to {file_path}."
        )

        write_data = _make_config_paths_relative(self.data, self.root.path)
        match file_path.suffix.lower():
            case ".yaml" | ".yml":
                write_yaml(file_path, write_data)
            case ".toml":
                write_toml(file_path, write_data)
            case _:
                raise ValueError(f"Unknown file extension: {file_path.suffix}")

    @hydromt_step
    def read(self, path: Optional[str] = None) -> None:
        """Read model config at <root>/{path}."""
        self._initialize(skip_read=True)

        if (
            self._data is not None
            and len(self._data) == 0
            and self._default_template_filename
        ):
            read_path = self.root.path / self._default_template_filename
        else:
            p = path or self._filename
            read_path = self.root.path / p

        if isfile(read_path):
            logger.info(f"Reading model config file from {read_path}.")
        else:
            logger.warning(
                f"No default model config was found at {read_path}. "
                "It wil be initialized as empty dictionary"
            )
            return

        # Always overwrite config when reading
        match read_path.suffix.lower():
            case ".yaml" | ".yml":
                self._data = read_yaml(read_path)
            case ".toml":
                self._data = read_toml(read_path)
            case _:
                raise ValueError(f"Unknown file extension: {read_path.suffix}")

    def set(self, key: str, value: Any):
        """Update the config dictionary at key(s) with values.

        Parameters
        ----------
        key : str
            a string with '.' indicating a new level: 'key1.key2' will translate
            to {"key1":{"key2": value}}
        value: Any
            the value to set the config to

        Examples
        --------
        ::

            >> self.set({'a': 1, 'b': {'c': {'d': 2}}})
            >> self.data
                {'a': 1, 'b': {'c': {'d': 2}}}
            >> self.set_value('a', 99)
            >> {'a': 99, 'b': {'c': {'d': 2}}}

            >> self.set_value('b.d.e', 24)
            >> {'a': 99, 'b': {'c': {'d': 24}}}
        """
        self._initialize()
        parts = key.split(".")
        num_parts = len(parts)
        current = cast(Dict[str, Any], self._data)
        for i, part in enumerate(parts):
            if part not in current or not isinstance(current[part], dict):
                current[part] = {}
            if i < num_parts - 1:
                current = current[part]
            else:
                current[part] = value

    def get_value(self, key: str, fallback=None, abs_path: bool = False) -> Any:
        """Get a config value at key(s).

        Parameters
        ----------
        args : tuple or string
            key can given as a string with '.' indicating a new level: ('key1.key2')
        fallback: any, optional
            Fallback value if key not found in config, by default None.
        abs_path: bool, optional
            If True return the absolute path relative to the model root,
            by default False.
            NOTE: this assumes the config is located in model root!

        Returns
        -------
        value : any type
            dictionary value

        Examples
        --------
        >> self.data = {'a': 1, 'b': {'c': {'d': 2}}}

        >> get_value('a')
        >> 1

        >> get_value('b.c.d')
        >> 2

        >> get_value('b.c')
        >> {'d': 2}

        """
        parts = key.split(".")
        num_parts = len(parts)
        current = self.data  # reads config at first call
        value = fallback
        for i, part in enumerate(parts):
            if i < num_parts - 1:
                current = current.get(part, {})
            else:
                value = current.get(part, fallback)

        if abs_path and isinstance(value, (str, Path)):
            value = Path(value)
            if not value.is_absolute():
                value = self.root.path / value

        return value

    @hydromt_step
    def create(
        self,
        template: Optional[Union[str, Path]] = None,
    ):
        """Create a new config file based on a template file.

        It the template is not provided, the default template will be used if available.
        Only yaml and toml files are supported.

        Parameters
        ----------
        template : str or Path, optional
            Path to a template config file, by default None

        Examples
        --------
        ::

            >> self.create()
            >> self.data
                {}

            >> self.create(template='path/to/template.yml')
            >> self.data
                {'a': 1, 'b': {'c': {'d': 2}}}
        """
        # Check if self.data is not empty
        if len(self.data) > 0:
            raise ValueError(
                "Model config already exists, cannot create new config."
                "Use ``update`` method to update the existing config."
            )

        if template is not None:
            prefix = "user-defined"
        elif self._default_template_filename is not None:
            template = self._default_template_filename
            prefix = "default"
        else:
            raise FileNotFoundError("No template file was provided.")

        if not isfile(template):
            raise FileNotFoundError(f"Template file not found: {template}")

        template = Path(template)
        # Here directly overwrite config with template
        logger.info(f"Creating model config from {prefix} template: {template}")
        match template.suffix.lower():
            case ".yaml" | ".yml":
                self._data = read_yaml(template)
            case ".toml":
                self._data = read_toml(template)
            case _:
                raise ValueError(f"Unknown file extension: {template.suffix}")

    @hydromt_step
    def update(self, data: Dict[str, Any]):
        """Set the config dictionary at key(s) with values.

        Parameters
        ----------
        data: Dict[str,Any]
            A dictionary with the values to be set. keys can be dotted like in
            :py:meth:`~hydromt.model.components.config.ConfigComponent.set_value`

        Examples
        --------
        Setting data as a nested dictionary::


            >> self.update({'a': 1, 'b': {'c': {'d': 2}}})
            >> self.data
            {'a': 1, 'b': {'c': {'d': 2}}}

        Setting data using dotted notation::

            >> self.update({'a.d.f.g': 1, 'b': {'c': {'d': 2}}})
            >> self.data
            {'a': {'d':{'f':{'g': 1}}}, 'b': {'c': {'d': 2}}}

        """
        if len(data) > 0:
            logger.debug("Setting model config options.")
        for k, v in data.items():
            self.set(k, v)

    def test_equal(self, other: ModelComponent) -> Tuple[bool, Dict[str, str]]:
        """Test if two components are equal.

        Parameters
        ----------
        other : ModelComponent
            The component to compare against.

        Returns
        -------
        tuple[bool, Dict[str, str]]
            True if the components are equal, and a dict with the associated errors per property checked.
        """
        eq, errors = super().test_equal(other)
        if not eq:
            return eq, errors
        other_config = cast(ConfigComponent, other)

        # for once python does the recursion for us
        if self.data == other_config.data:
            return True, {}
        else:
            return False, {"config": "Configs are not equal"}



"""Generic driver for reading and writing DataFrames."""

import logging
from abc import ABC, abstractmethod
from typing import Any, ClassVar

import pandas as pd

from hydromt.data_catalog.drivers import BaseDriver
from hydromt.error import NoDataStrategy
from hydromt.typing import (
    SourceMetadata,
    StrPath,
    TimeRange,
    Variables,
)

logger = logging.getLogger(__name__)


class DataFrameDriver(BaseDriver, ABC):
    """Abstract Driver to read DataFrames."""

    supports_writing: ClassVar[bool] = False

    @abstractmethod
    def read(
        self,
        uris: list[str],
        *,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
        kwargs_for_open: dict[str, Any] | None = None,
        variables: Variables | None = None,
        time_range: TimeRange | None = None,
        metadata: SourceMetadata | None = None,
    ) -> pd.DataFrame:
        """Read in any compatible data source to a pandas `DataFrame`."""
        ...

    def write(
        self,
        path: StrPath,
        df: pd.DataFrame,
        **kwargs,
    ) -> str:
        """
        Write out a DataFrame to file.

        Not all drivers should have a write function, so this method is not
        abstract.

        args:
        """
        raise NotImplementedError(
            f"Writing using driver '{self.name}' is not supported."
        )



# -*- coding: utf-8 -*-
"""Utils for parsing cli options and arguments."""

import json
from os.path import isfile
from typing import Any, Dict

import click

__all__ = ["parse_json"]


### CLI callback methods ###
def parse_json(_ctx: click.Context, _param, value: str) -> Dict[str, Any]:
    """Parse json from object or file.

    If the object passed is a path pointing to a file, load it's contents and parse it.
    Otherwise attempt to parse the object as JSON itself.
    """
    if isfile(value):
        with open(value, "r") as f:
            kwargs = json.load(f)
    else:
        if value.strip("{").startswith("'"):
            value = value.replace("'", '"')
        try:
            kwargs = json.loads(value)
        except json.JSONDecodeError:
            raise ValueError(f'Could not decode JSON "{value}"')
    return kwargs



import sys
from typing import List

from packaging.version import Version

__all__: List[str] = []

HAS_GCSFS = False
HAS_GDAL = False
HAS_OPENPYXL = False
HAS_PYET = False
HAS_S3FS = False

try:
    import gcsfs

    HAS_GCSFS = True
except ImportError:
    pass

try:
    from osgeo import gdal

    HAS_GDAL = True
except ImportError:
    pass

try:
    import openpyxl

    HAS_OPENPYXL = True
except ImportError:
    pass

try:
    import pyet

    HAS_PYET = True
except ModuleNotFoundError:
    pass

try:
    import s3fs

    HAS_S3FS = True
except ImportError:
    pass

# entrypoints in standard library only compatible from 3.10 onwards
py_version = sys.version_info
if py_version[0] >= 3 and py_version[1] >= 10:
    from importlib.metadata import entry_points, Distribution, EntryPoint, EntryPoints  # noqa: I001
else:
    from importlib_metadata import entry_points, Distribution, EntryPoint, EntryPoints  # noqa: I001



"""parse a region from a dict. See parse_region for information on usage."""

import logging
from os.path import isdir, isfile
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import geopandas as gpd
import numpy as np
import xarray as xr
import xugrid as xu
from pyproj import CRS
from shapely import box

from hydromt.data_catalog import DataCatalog
from hydromt.data_catalog.sources import RasterDatasetSource
from hydromt.error import NoDataStrategy
from hydromt.gis import parse_crs
from hydromt.model.processes.basin_mask import get_basin_geometry
from hydromt.plugins import PLUGINS
from hydromt.typing.type_def import StrPath

if TYPE_CHECKING:
    from hydromt.model.model import Model


logger = logging.getLogger(__name__)

__all__ = [
    "parse_region_basin",
    "parse_region_bbox",
    "parse_region_geom",
    "parse_region_grid",
    "parse_region_other_model",
    "parse_region_mesh",
]


def parse_region_basin(
    region: dict,
    *,
    data_catalog: DataCatalog,
    hydrography_path: StrPath,
    basin_index_path: Optional[StrPath] = None,
) -> gpd.GeoDataFrame:
    """Parse a basin /subbasin / interbasin region and return the GeoDataFrame.

    Parameters
    ----------
    region : dict
        Dictionary describing region of interest.
        Entire basin can be defined based on an ID, one or multiple point location
        (x, y), or a region of interest (bounding box or geometry) for which the
        basin IDs are looked up. The basins withint the area of interest can be further
        filtered to only include basins with their outlet within the area of interest
        ('outlets': true) of stream threshold arguments (e.g.: 'uparea': 1000).

        Common use-cases include:
        * {'basin': ID}
        * {'basin': [ID1, ID2, ..]}
        * {'basin': [x, y]}
        * {'basin': [[x1, x2, ..], [y1, y2, ..]]}
        * {'basin': /path/to/point_geometry}
        * {'basin': [xmin, ymin, xmax, ymax]}
        * {'basin': [xmin, ymin, xmax, ymax], 'outlets': true}
        * {'basin': [xmin, ymin, xmax, ymax], '<variable>': threshold}

        Subbasins are defined by its outlet locations and include all area upstream
        from these points. The outlet locations can be passed as xy coordinate pairs,
        but also derived from the most downstream cell(s) within a area of interest
        defined by a bounding box or geometry, optionally refined by stream threshold
        arguments.

        The method can be speed up by providing an additional ``bounds`` argument which
        should contain all upstream cell. If cells upstream of the subbasin are not
        within the provide bounds a warning will be raised. Common use-cases include:

        * {'subbasin': [x, y], '<variable>': threshold}
        * ::

            {
                'subbasin': [[x1, x2, ..], [y1, y2, ..]],
                '<variable>': threshold, 'bounds': [xmin, ymin, xmax, ymax]
            }

        * {'subbasin': /path/to/point_geometry, '<variable>': threshold}
        * {'subbasin': [xmin, ymin, xmax, ymax], '<variable>': threshold}
        * {'subbasin': /path/to/polygon_geometry, '<variable>': threshold}

        Interbasins are similar to subbasins but are bounded by a bounding box or
        geometry and do not include all upstream area. Common use-cases include:

        * {'interbasin': [xmin, ymin, xmax, ymax], '<variable>': threshold}
        * {'interbasin': [xmin, ymin, xmax, ymax], 'xy': [x, y]}
        * {'interbasin': /path/to/polygon_geometry, 'outlets': true}
    data_catalog : DataCatalog
        DataCatalog object containing the data sources.
    hydrography_fn : strPath
        Path of the hydrography raster dataset in the data catalog.
    basin_index_path : StrPath, optional
        Path of the basin index raster dataset in the data catalog.
    """
    var_thresh_kwargs = region.copy()
    kind = next(iter(region))
    value0 = var_thresh_kwargs.pop(kind)

    _assert_parse_key(kind, "basin", "interbasin", "subbasin")

    kwargs = _parse_region_value(value0, data_catalog=data_catalog)
    kwargs.update(var_thresh_kwargs)

    if kind == "basin":
        expected_keys = ["basid", "geom", "bbox", "xy"]
    elif kind == "interbasin":
        expected_keys = ["geom", "bbox"]
    else:
        expected_keys = ["geom", "bbox", "xy"]
    _assert_parsed_values(
        key=next(iter(kwargs)), region_value=value0, kind=kind, expected=expected_keys
    )
    kwargs_str = dict()
    for k, v in kwargs.items():
        if isinstance(v, gpd.GeoDataFrame):
            v = f"GeoDataFrame {v.total_bounds} (crs = {v.crs})"
        kwargs_str.update({k: v})
    logger.debug(f"Parsed region (kind={kind}): {str(kwargs_str)}")

    ds_org = data_catalog.get_rasterdataset(hydrography_path)
    if "bounds" not in kwargs and basin_index_path is not None:
        kwargs.update(basin_index=data_catalog.get_source(str(basin_index_path)))
    # get basin geometry
    geom, _ = get_basin_geometry(ds=ds_org, kind=kind, **kwargs)

    return geom


def parse_region_bbox(region: dict, *, crs: int = 4326) -> gpd.GeoDataFrame:
    """Parse a region of kind bbox and return the GeoDataFrame.

    Parameters
    ----------
    region : dict
        Dictionary describing region of interest.
        For an exact clip of the region: { 'bbox': [xmin, ymin, xmax, ymax] }
    crs : CRS, optional
        CRS of the bounding box coordinates. By default EPSG 4326.
    """
    kwargs = region.copy()
    kind = next(iter(region))
    value0 = kwargs.pop(kind)

    _assert_parse_key(kind, "bbox")

    kwargs.update(_parse_region_value(value0, data_catalog=None))

    _assert_parsed_values(
        key=next(iter(kwargs)), region_value=value0, kind="bbox", expected=["bbox"]
    )
    logger.debug(f"Parsed region (kind={kind}): {str(kwargs)}")

    bbox = kwargs["bbox"]
    geom = gpd.GeoDataFrame(geometry=[box(*bbox)], crs=crs)
    return geom


def parse_region_geom(
    region: dict,
    *,
    crs: Optional[int] = None,
    data_catalog: Optional[DataCatalog] = None,
) -> gpd.GeoDataFrame:
    """Parse a region and return the GeoDataFrame.

    Parameters
    ----------
    region : dict
        Dictionary describing region of interest.
        For an exact clip of the region:
        * {'geom': /path/to/polygon_geometry}
    crs : CRS, optional
        CRS of the geometry in case it cannot be derived from the geometry itself.
    data_catalog : DataCatalog, optional
        DataCatalog object containing the data sources.
    """
    kwargs = region.copy()
    kind = next(iter(region))
    value0 = kwargs.pop(kind)

    _assert_parse_key(kind, "geom")

    kwargs.update(_parse_region_value(value0, data_catalog=data_catalog))

    _assert_parsed_values(
        key=next(iter(kwargs)), region_value=value0, kind="geom", expected=["geom"]
    )
    kwargs_str = dict()
    for k, v in kwargs.items():
        if isinstance(v, gpd.GeoDataFrame):
            v = f"GeoDataFrame {v.total_bounds} (crs = {v.crs})"
        kwargs_str.update({k: v})
    logger.debug(f"Parsed region (kind={kind}): {str(kwargs_str)}")

    geom = kwargs["geom"]
    if geom.crs is None:
        logger.debug(f'Model region "geom" has no CRS, setting to {crs}')
        _update_crs(geom, crs)

    return geom


def parse_region_grid(
    region: dict,
    *,
    data_catalog: Optional[DataCatalog],
) -> Union[xr.DataArray, xr.Dataset]:
    """Parse a region of kind grid and return the corresponding xarray object.

    Note that additional arguments from the region can be passed to the
    ``driver_kwargs`` argument of the hydromt.data_catalog.DataCatalog.get_rasterdataset
    method.

    Parameters
    ----------
    region : dict
        Dictionary describing region of interest.
        For a region based on a grid:

        * {'grid': /path/to/grid}
    data_catalog : DataCatalog
        DataCatalog object containing the data sources.

    Returns
    -------
    xr.DataArray or xr.Dataset
        The parsed xarray object.
    """
    kwargs = region.copy()
    kind = next(iter(region))
    value0 = kwargs.pop(kind)

    _assert_parse_key(kind, "grid")

    kwargs_str = dict()
    for k, v in kwargs.items():
        if isinstance(v, xr.DataArray):
            v = f"DataArray {v.raster.bounds} (crs = {v.raster.crs})"
        kwargs_str.update({k: v})
    logger.debug(f"Parsed region (kind={kind}): {str(kwargs_str)}")

    data_catalog = data_catalog or DataCatalog()
    da = data_catalog.get_rasterdataset(
        value0,
        handle_nodata=NoDataStrategy.RAISE,
        single_var_as_array=True,
        driver={"name": RasterDatasetSource._fallback_driver_read, "options": kwargs},
    )

    return da


def parse_region_other_model(region: dict) -> "Model":
    """Parse a region with a model path and return that whole Model in read mode.

    Parameters
    ----------
    region : dict
        Dictionary describing region of interest.
        For a region based of another models grid:
        * {'<model_name>': root}
    """
    kwargs = region.copy()
    kind = next(iter(region))
    value0 = kwargs.pop(kind)

    _assert_parse_key(kind, *PLUGINS.model_plugins.keys())
    logger.debug(f"Parsed region (kind={kind}): {str(kwargs)}")

    model_class = PLUGINS.model_plugins[kind]
    return model_class(root=value0, mode="r")


def parse_region_mesh(region: dict) -> xu.UgridDataset:
    """Parse a region with a mesh path and return that mesh in read mode.

    Parameters
    ----------
    region : dict
        Dictionary describing region of interest.
        For a region based of a mesh grid of a mesh file:
        * {'mesh': /path/to/mesh}
        * {'mesh': UgridDataArray}
        * {'mesh': UgridDataset}
        * {'mesh': Ugrid1d}
        * {'mesh': Ugrid2d}
    """
    kwargs = region.copy()
    kind = next(iter(region))
    value0 = kwargs.pop(kind)

    _assert_parse_key(kind, "mesh")

    if isinstance(value0, (str, Path)) and isfile(value0):
        return xu.open_dataset(value0)
    elif isinstance(value0, (xu.UgridDataset, xu.UgridDataArray)):
        return value0
    elif isinstance(value0, (xu.Ugrid1d, xu.Ugrid2d)):
        return xu.UgridDataset(value0.to_dataset(optional_attributes=True))
    else:
        raise ValueError(
            f"Unrecognized type {type(value0)}.Should be a path or xugrid object."
        )


def _update_crs(
    geom: gpd.GeoDataFrame, crs: Optional[Union[CRS, int]]
) -> Optional[gpd.GeoDataFrame]:
    if crs is not None:
        crs = parse_crs(crs, bbox=geom.total_bounds)
        return geom.to_crs(crs)
    return geom


def _parse_region_value(
    value: Any, *, data_catalog: Optional[DataCatalog]
) -> Dict[str, Any]:
    kwarg: Dict[str, Any] = {}
    if isinstance(value, np.ndarray):
        value = value.tolist()  # array to list

    if isinstance(value, list):
        if np.all([isinstance(p0, int) and abs(p0) > 180 for p0 in value]):  # all int
            kwarg = dict(basid=value)
        elif len(value) == 4:  # 4 floats
            kwarg = dict(bbox=value)
        elif len(value) == 2:  # 2 floats
            kwarg = dict(xy=value)
    elif isinstance(value, tuple) and len(value) == 2:  # tuple of x and y coords
        kwarg = dict(xy=value)
    elif isinstance(value, int):  # single int
        kwarg = dict(basid=value)
    elif isinstance(value, (str, Path)) and isdir(value):
        kwarg = dict(root=value)
    elif isinstance(value, (str, Path)):
        data_catalog = data_catalog or DataCatalog()
        geom = data_catalog.get_geodataframe(value)
        kwarg = dict(geom=geom)
    elif isinstance(value, gpd.GeoDataFrame):  # geometry
        kwarg = dict(geom=value)
    else:
        raise ValueError(f"Region value {value} not understood.")

    if "geom" in kwarg and np.all(kwarg["geom"].geometry.type == "Point"):
        xy = (
            kwarg["geom"].geometry.x.values,
            kwarg["geom"].geometry.y.values,
        )
        kwarg = dict(xy=xy)
    return kwarg


def _assert_parse_key(key: str, *expected: str) -> None:
    if key not in expected:
        raise KeyError(f"Expected key in '{', '.join(expected)}' but got '{key}'.")


def _assert_parsed_values(
    *, key: str, region_value: Any, kind: str, expected: List[str]
) -> None:
    if key not in expected:
        raise ValueError(
            f"Region value '{region_value}' for kind={kind} not understood, "
            f"provide one of {','.join(expected)}"
        )



"""Geoms component."""

import logging
from glob import glob
from typing import TYPE_CHECKING, Dict, Optional, Tuple, Union, cast

import geopandas as gpd
import numpy as np
from geopandas import GeoDataFrame, GeoSeries
from geopandas.testing import assert_geodataframe_equal
from shapely.geometry import box

from hydromt._utils.naming_convention import _expand_uri_placeholders
from hydromt.model.components.base import ModelComponent
from hydromt.model.components.spatial import SpatialModelComponent
from hydromt.model.steps import hydromt_step

if TYPE_CHECKING:
    from hydromt.model.model import Model

logger = logging.getLogger(__name__)


class GeomsComponent(SpatialModelComponent):
    """A component to manage geo-spatial geometries.

    It contains a dictionary of geopandas GeoDataFrames.
    """

    def __init__(
        self,
        model: "Model",
        *,
        filename: str = "geoms/{name}.geojson",
        region_component: Optional[str] = None,
        region_filename: str = "geoms/geoms_region.geojson",
    ):
        """Initialize a GeomsComponent.

        Parameters
        ----------
        model: Model
            HydroMT model instance
        filename: str
            The path to use for reading and writing of component data by default.
            by default "geoms/{name}.geojson" ie one file per geodataframe in the data dictionary.
        region_component: str, optional
            The name of the region component to use as reference for this component's
            region. If None, the region will be set to the union of all geometries in
            the data dictionary.
        region_filename: str
            The path to use for writing the region data to a file. By default
            "geoms/geoms_region.geojson".
        """
        self._data: Optional[Dict[str, Union[GeoDataFrame, GeoSeries]]] = None
        self._filename: str = filename
        super().__init__(
            model=model,
            region_component=region_component,
            region_filename=region_filename,
        )

    @property
    def data(self) -> Dict[str, Union[GeoDataFrame, GeoSeries]]:
        """Model geometries.

        Return dict of geopandas.GeoDataFrame or geopandas.GeoSeries
        """
        if self._data is None:
            self._initialize()

        assert self._data is not None
        return self._data

    def _initialize(self, skip_read=False) -> None:
        """Initialize geoms."""
        if self._data is None:
            self._data = dict()
            if self.root.is_reading_mode() and not skip_read:
                self.read()

    @property
    def _region_data(self) -> Optional[GeoDataFrame]:
        # Use the total bounds of all geometries as region
        if len(self.data) == 0:
            return None
        bounds = np.column_stack([geom.total_bounds for geom in self.data.values()])
        total_bounds = (
            bounds[0, :].min(),
            bounds[1, :].min(),
            bounds[2, :].max(),
            bounds[3, :].max(),
        )
        region = gpd.GeoDataFrame(geometry=[box(*total_bounds)], crs=self.model.crs)

        return region

    def set(self, geom: Union[GeoDataFrame, GeoSeries], name: str):
        """Add data to the geom component.

        Arguments
        ---------
        geom: geopandas.GeoDataFrame or geopandas.GeoSeries
            New geometry data to add
        name: str
            Geometry name.
        """
        self._initialize()
        assert self._data is not None
        if name in self._data:
            logger.warning(f"Replacing geom: {name}")

        if isinstance(geom, GeoSeries):
            geom = cast(GeoDataFrame, geom.to_frame())

        # Verify if a geom is set to model crs and if not sets geom to model crs
        model_crs = self.model.crs
        if model_crs and model_crs != geom.crs:
            geom.to_crs(model_crs.to_epsg(), inplace=True)
        self._data[name] = geom

    @hydromt_step
    def read(self, filename: Optional[str] = None, **kwargs) -> None:
        r"""Read model geometries files at <root>/<filename>.

        key-word arguments are passed to :py:func:`geopandas.read_file`

        Parameters
        ----------
        filename : str, optional
            filename relative to model root. should contain a {name} placeholder
            which will be used to determine the names/keys of the geometries.
            if None, the path that was provided at init will be used.
        **kwargs:
            Additional keyword arguments that are passed to the
            `geopandas.read_file` function.
        """
        self.root._assert_read_mode()
        self._initialize(skip_read=True)
        f = filename or self._filename
        read_path = self.root.path / f
        path_glob, _, regex = _expand_uri_placeholders(str(read_path))
        paths = glob(path_glob)
        for p in paths:
            name = ".".join(regex.match(p).groups())  # type: ignore
            geom = cast(GeoDataFrame, gpd.read_file(p, **kwargs))
            logger.debug(f"Reading model file {name} at {p}.")

            self.set(geom=geom, name=name)

    @hydromt_step
    def write(
        self,
        filename: Optional[str] = None,
        *,
        to_wgs84: bool = False,
        **kwargs,
    ) -> None:
        r"""Write model geometries to a vector file (by default GeoJSON) at <root>/<filename>.

        key-word arguments are passed to :py:meth:`geopandas.GeoDataFrame.to_file`

        Parameters
        ----------
        filename : str, optional
            filename relative to model root. should contain a {name} placeholder
            which will be used to determine the names/keys of the geometries.
            if None, the path that was provided at init will be used.
            Can be a relative path.
        to_wgs84: bool, optional
            If True, the geoms will be reprojected to WGS84(EPSG:4326) before they are written.
        **kwargs:
            Additional keyword arguments that are passed to the
            `geopandas.to_file` function.
        """
        self.root._assert_write_mode()

        if len(self.data) == 0:
            logger.info(
                f"{self.model.name}.{self.name_in_model}: No geoms data found, skip writing."
            )
            return

        filename = filename or self._filename

        for name, gdf in self.data.items():
            if len(gdf) == 0:
                logger.warning(
                    f"{self.model.name}.{self.name_in_model}: {name} is empty. Skipping..."
                )
                continue

            write_path = self.root.path / filename.format(name=name)
            write_path.parent.mkdir(parents=True, exist_ok=True)
            logger.info(
                f"{self.model.name}.{self.name_in_model}: Writing geoms to {write_path}."
            )

            if to_wgs84 and (
                kwargs.get("driver") == "GeoJSON"
                or write_path.suffix.lower() == ".geojson"
            ):
                gdf.to_crs(epsg=4326, inplace=True)

            gdf.to_file(write_path, **kwargs)

    def test_equal(self, other: ModelComponent) -> Tuple[bool, Dict[str, str]]:
        """Test if two GeomsComponents are equal.

        Parameters
        ----------
        other: GeomsComponent
            The other GeomsComponent to compare with.

        Returns
        -------
        tuple[bool, dict[str, str]]
            True if the components are equal, and a dict with the associated errors per property checked.
        """
        eq, errors = super().test_equal(other)
        if not eq:
            return eq, errors
        other_geoms = cast(GeomsComponent, other)
        for name, gdf in self.data.items():
            if name not in other_geoms.data:
                errors[name] = "Geom not found in other component."
            try:
                assert_geodataframe_equal(
                    gdf,
                    other_geoms.data[name],
                    check_like=True,
                    check_less_precise=True,
                )
            except AssertionError as e:
                errors[name] = str(e)

        return len(errors) == 0, errors
